import asyncio
import base64
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Any, Callable, Optional
import uuid

from opentelemetry import trace

from meshagent.api import (
    Element,
    MeshDocument,
    Participant,
    RemoteParticipant,
    RoomClient,
    RoomException,
)
from meshagent.tools import Toolkit, tool

from meshagent.agents.agent import AgentSessionContext
from meshagent.agents.thread_schema import thread_schema

tracer = trace.get_tracer("meshagent.thread_adapter")
logger = logging.getLogger("thread_adapter")


def default_format_message(*, user_name: str, message: str, iso_timestamp: str) -> str:
    return f"{user_name} said at {iso_timestamp}: {message}"


class ThreadAdapter(ABC):
    def __init__(
        self,
        *,
        room: RoomClient,
        path: str,
        format_message: Optional[Callable] = None,
        max_append_message_count: int = 25,
    ):
        self._room = room
        self._thread_path = path
        self._processor_task: Optional[asyncio.Task] = None
        self._llm_messages: asyncio.Queue = asyncio.Queue()
        self._thread: Optional[MeshDocument] = None
        self._format_message = format_message or default_format_message
        self._max_append_message_count = max_append_message_count

    async def start(self) -> None:
        self._thread = await self._room.sync.open(
            path=self._thread_path,
            schema=thread_schema,
        )
        self._ensure_members_element()
        self._ensure_messages_element()
        self._processor_task = asyncio.create_task(self._process_llm_events())

    async def __aenter__(self) -> "ThreadAdapter":
        await self.start()
        return self

    @property
    def thread(self) -> Optional[MeshDocument]:
        return self._thread

    @property
    def path(self) -> str:
        return self._thread_path

    def format_message(
        self,
        *,
        user_name: str,
        message: str,
        iso_timestamp: str,
    ) -> str:
        return self._format_message(
            user_name=user_name,
            message=message,
            iso_timestamp=iso_timestamp,
        )

    async def stop(self) -> None:
        if self._processor_task is not None and not self._processor_task.done():
            # Give the processor a bounded window to consume pending events
            # before we shut down the queue.
            drain_deadline = asyncio.get_running_loop().time() + 2.0
            while not self._llm_messages.empty():
                if self._processor_task.done():
                    break
                if asyncio.get_running_loop().time() >= drain_deadline:
                    break
                await asyncio.sleep(0.01)

        self._llm_messages.shutdown()
        if self._processor_task is not None:
            await self._processor_task
            self._processor_task = None

        if self._thread is not None:
            final_state: bytes | None = None
            try:
                state = self._thread.get_state()
                if isinstance(state, bytes) and len(state) > 0:
                    final_state = state
            except Exception as ex:
                logger.warning(
                    "unable to collect final thread state for %s",
                    self._thread_path,
                    exc_info=ex,
                )

            if final_state is not None:
                try:
                    encoded_state = base64.standard_b64encode(final_state)
                    await self._room.sync.sync(
                        path=self._thread_path,
                        data=encoded_state,
                    )
                except Exception as ex:
                    logger.warning(
                        "unable to flush final thread state for %s",
                        self._thread_path,
                        exc_info=ex,
                    )

            # TODO: Wait for pending changes to sync
            await asyncio.sleep(3)
            await self._room.sync.close(path=self._thread_path)
            self._thread = None

    async def __aexit__(self, exc_type, exc, tb) -> None:
        del exc_type
        del exc
        del tb
        await self.stop()

    def push(self, *, event: dict) -> None:
        try:
            self._llm_messages.put_nowait(event)
        except asyncio.QueueShutDown:
            logger.debug("dropping thread adapter event after queue shutdown")

    async def clear_thread(self) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages = self._ensure_messages_element()

        for child in list(messages.get_children()):
            child.delete()

    @abstractmethod
    async def handle_custom_event(
        self,
        *,
        messages: Element,
        event: dict,
    ) -> None: ...

    @abstractmethod
    async def _process_llm_events(self) -> None: ...

    def make_toolkit(self) -> Toolkit:
        return Toolkit(
            name="search",
            description="tools for searching conversation history",
            tools=[
                self.grep_tool,
                self.get_message_range,
                self.count_tool,
            ],
        )

    @tool(
        name="get_message_range",
        description="gets a range of messages, index 0 is the first message in the conversation",
    )
    def get_message_range(self, *, start: int, end: int) -> str:
        all_items = self._thread.root.get_children_by_tag_name("messages")[
            0
        ].get_children()
        messages = [
            item
            for item in all_items
            if isinstance(item, Element) and item.tag_name == "message"
        ]

        elements = messages[start:end]

        if len(elements) == 0:
            return "no messages were found within the specified range"

        response = "matching messages:\n"

        for element in elements:
            response = response + self._format_message(
                user_name=element["author_name"],
                message=element["text"],
                iso_timestamp=element["created_at"],
            )

        return response

    @tool(
        name="count_current_thread_messages",
        description="return the number of messages in the current thread (including those outside the context window)",
    )
    def count_tool(
        self,
        *,
        pattern: str,
        ignore_case: bool,
        messages_before: int,
        messages_after: int,
    ) -> str:
        del pattern
        del ignore_case
        del messages_before
        del messages_after

        messages = self._thread.root.get_children_by_tag_name("messages")[0]
        count = len(
            [
                item
                for item in messages.get_children()
                if isinstance(item, Element) and item.tag_name == "message"
            ]
        )
        return f"{count}"

    @tool(
        name="grep_current_thread",
        description="search the current thread for text, includes messages outside the current context window",
    )
    def grep_tool(
        self,
        *,
        pattern: str,
        ignore_case: bool,
        messages_before: int,
        messages_after: int,
    ) -> str:
        messages = self._thread.root.get_children_by_tag_name("messages")[0]
        elements = [
            item
            for item in messages.grep(
                pattern,
                ignore_case=ignore_case,
                before=messages_before,
                after=messages_after,
            )
            if isinstance(item, Element) and item.tag_name == "message"
        ]
        if len(elements) == 0:
            return "no messages were found with the specified pattern"

        response = "matching messages:\n"

        for element in elements:
            response = response + self._format_message(
                user_name=element["author_name"],
                message=element["text"],
                iso_timestamp=element["created_at"],
            )

        return response

    def append_messages(self, *, context: AgentSessionContext) -> None:
        doc_messages = None

        for prop in self._thread.root.get_children():
            if prop.tag_name == "messages":
                doc_messages = prop

                messages = [
                    element
                    for element in doc_messages.get_children()
                    if isinstance(element, Element) and element.tag_name == "message"
                ]
                if len(messages) > self._max_append_message_count:
                    first_message = len(messages) - self._max_append_message_count
                    messages = messages[first_message:]
                    context.append_assistant_message(
                        f"there are more messages outside the current context window, the index of the first message loaded is {first_message}"
                    )

                for element in messages:
                    if isinstance(element, Element) and element.tag_name == "message":
                        msg = element["text"]
                        if element[
                            "author_name"
                        ] == self._room.local_participant.get_attribute("name"):
                            context.append_assistant_message(msg)
                        else:
                            context.append_user_message(
                                self._format_message(
                                    user_name=element["author_name"],
                                    message=element["text"],
                                    iso_timestamp=element["created_at"],
                                )
                            )

                        for child in element.get_children():
                            if child.tag_name == "file":
                                context.append_assistant_message(
                                    f"the user attached a file with the path '{child.get_attribute('path')}'"
                                )

                break

        if doc_messages is None:
            doc_messages = self._ensure_messages_element()

    def _ensure_members_element(self) -> Element:
        if self._thread is None:
            raise RoomException("thread was not opened")

        members = self._thread.root.get_children_by_tag_name("members")
        if len(members) > 0:
            return members[0]

        return self._thread.root.append_child(tag_name="members")

    def _ensure_messages_element(self) -> Element:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages = self._thread.root.get_children_by_tag_name("messages")
        if len(messages) > 0:
            return messages[0]

        return self._thread.root.append_child(tag_name="messages")

    def ensure_member(self, *, participant: Participant | str) -> None:
        if isinstance(participant, Participant):
            participant_name = participant.get_attribute("name")
        else:
            participant_name = participant

        if not isinstance(participant_name, str):
            return

        normalized_name = participant_name.strip()
        if normalized_name == "":
            return

        members = self._ensure_members_element()
        for member in members.get_children():
            if member.tag_name != "member":
                continue

            if member.get_attribute("name") == normalized_name:
                return

        members.append_child(
            tag_name="member",
            attributes={"name": normalized_name},
        )

    def write_text_message(
        self,
        *,
        text: str,
        participant: Participant | str,
        message_id: Optional[str] = None,
        turn_id: Optional[str] = None,
        attachments: Optional[list[dict[str, Any]]] = None,
        role: Optional[str] = None,
    ) -> None:
        self.ensure_member(participant=participant)
        doc_messages = self._ensure_messages_element()

        author_name = ""
        if isinstance(participant, Participant):
            participant_name = participant.get_attribute("name")
            if isinstance(participant_name, str):
                author_name = participant_name
        else:
            author_name = participant

        attributes: dict[str, Any] = {
            "text": text,
            "created_at": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            "author_name": author_name,
        }

        normalized_role = role.strip() if isinstance(role, str) else ""
        if normalized_role == "" and isinstance(participant, RemoteParticipant):
            normalized_role = participant.role.strip()
        if normalized_role == "" and participant is self._room.local_participant:
            normalized_role = "agent"
        if normalized_role != "":
            attributes["role"] = normalized_role

        if isinstance(message_id, str) and message_id.strip() != "":
            attributes["id"] = message_id.strip()
        if isinstance(turn_id, str) and turn_id.strip() != "":
            attributes["turn_id"] = turn_id.strip()

        message = doc_messages.append_child(
            tag_name="message",
            attributes=attributes,
        )

        if attachments is None:
            return

        for attachment in attachments:
            if not isinstance(attachment, dict):
                continue
            path = attachment.get("path")
            if not isinstance(path, str):
                continue
            normalized_path = path.strip()
            if normalized_path == "":
                continue
            message.append_child(
                tag_name="file",
                attributes={"path": normalized_path},
            )

    def write_image(
        self,
        *,
        message_id: Optional[str],
        turn_id: Optional[str] = None,
        image_id: Optional[str] = None,
        mime_type: Optional[str] = None,
        created_at: Optional[str] = None,
        created_by: Optional[str] = None,
        width: Optional[int | float] = None,
        height: Optional[int | float] = None,
        status: Optional[str] = None,
        status_detail: Optional[str] = None,
    ) -> str:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages: Element = self._thread.root.get_children_by_tag_name("messages")[0]

        resolved_message_id = (
            message_id
            if isinstance(message_id, str) and message_id.strip() != ""
            else str(uuid.uuid4())
        )

        message = None
        for child in messages.get_children():
            if (
                child.tag_name == "message"
                and child.get_attribute("id") == resolved_message_id
            ):
                message = child
                break

        if message is None:
            author_name = (
                created_by
                if isinstance(created_by, str) and created_by.strip() != ""
                else self._room.local_participant.get_attribute("name")
            )
            if not isinstance(author_name, str):
                author_name = ""

            message_attributes: dict[str, Any] = {
                "id": resolved_message_id,
                "text": "",
                "created_at": datetime.now(timezone.utc)
                .isoformat()
                .replace("+00:00", "Z"),
                "author_name": author_name,
            }
            if isinstance(turn_id, str) and turn_id.strip() != "":
                message_attributes["turn_id"] = turn_id.strip()

            message = messages.append_child(
                tag_name="message",
                attributes=message_attributes,
            )
        elif isinstance(turn_id, str) and turn_id.strip() != "":
            message.set_attribute("turn_id", turn_id.strip())

        image = None
        for child in message.get_children():
            if child.tag_name == "image":
                image = child
                break

        normalized_width: Optional[int] = None
        if isinstance(width, (int, float)):
            width_int = int(width)
            if width_int > 0:
                normalized_width = width_int

        normalized_height: Optional[int] = None
        if isinstance(height, (int, float)):
            height_int = int(height)
            if height_int > 0:
                normalized_height = height_int

        image_attributes: dict[str, str | int] = {}
        if isinstance(image_id, str) and image_id.strip() != "":
            image_attributes["id"] = image_id
        if isinstance(mime_type, str) and mime_type.strip() != "":
            image_attributes["mime_type"] = mime_type
        if isinstance(created_at, str) and created_at.strip() != "":
            image_attributes["created_at"] = created_at
        if isinstance(created_by, str) and created_by.strip() != "":
            image_attributes["created_by"] = created_by
        if normalized_width is not None:
            image_attributes["width"] = normalized_width
        if normalized_height is not None:
            image_attributes["height"] = normalized_height
        if isinstance(status, str) and status.strip() != "":
            image_attributes["status"] = status.strip()
        if isinstance(status_detail, str) and status_detail.strip() != "":
            image_attributes["status_detail"] = status_detail.strip()

        if image is None:
            message.append_child(tag_name="image", attributes=image_attributes)
            return resolved_message_id

        for key, value in image_attributes.items():
            image.set_attribute(key, value)

        return resolved_message_id

    def set_message_turn_id(
        self,
        *,
        message_id: str,
        turn_id: str,
    ) -> bool:
        normalized_message_id = message_id.strip()
        normalized_turn_id = turn_id.strip()
        if normalized_message_id == "" or normalized_turn_id == "":
            return False

        messages = self._ensure_messages_element()
        for child in messages.get_children():
            if child.tag_name != "message":
                continue
            if child.get_attribute("id") != normalized_message_id:
                continue
            child.set_attribute("turn_id", normalized_turn_id)
            return True

        return False


# Backwards-compatible import path for existing callers.
def response_event_to_agent_event(event: dict) -> Optional[dict]:
    from .responses_thread_adapter import response_event_to_agent_event as _convert

    return _convert(event)



import base64
import logging
import posixpath
import re
import uuid
from datetime import datetime, timezone
from typing import Literal, Optional

from meshagent.api.schema_util import merge
from meshagent.api import Element, MeshDocument
from meshagent.tools import Toolkit

from .adapter import LLMAdapter
from .context import TaskContext
from .task_runner import TaskRunner
from .thread_schema import thread_list_schema
from .thread_adapter import ThreadAdapter
from .responses_thread_adapter import ResponsesThreadAdapter

logger = logging.getLogger(__name__)

ThreadingMode = Literal["auto", "manual", "none"]

DEFAULT_THREAD_NAME_RULES = [
    "generate a concise topic name for storing this task in a thread",
    "return only a thread_name value suitable for a file name",
    "thread_name should be 2-6 words, lowercase, and topic-focused",
    "do not include slashes or a .thread extension",
]


class ThreadedTaskRunner(TaskRunner):
    @staticmethod
    def resolve_threading_mode(
        *,
        threading_mode: Optional[ThreadingMode],
        input_path: bool,
    ) -> ThreadingMode:
        if threading_mode is None:
            return "manual" if input_path else "none"
        return threading_mode

    @staticmethod
    def with_manual_thread_path_schema(
        *,
        input_schema: dict,
        threading_mode: ThreadingMode,
    ) -> dict:
        if threading_mode != "manual":
            return input_schema
        return merge(
            schema=input_schema,
            additional_properties={"path": {"type": ["string", "null"]}},
        )

    def __init__(
        self,
        *,
        name=None,
        title=None,
        description=None,
        requires=None,
        supports_tools: Optional[bool] = None,
        input_schema: dict,
        output_schema: Optional[dict] = None,
        annotations: Optional[list[str]] = None,
        toolkits: Optional[list[Toolkit]] = None,
        input_path: bool = False,
        threading_mode: Optional[ThreadingMode] = None,
        thread_dir: str = ".threads",
        thread_name_rules: Optional[list[str]] = None,
        thread_name_adapter: Optional[LLMAdapter] = None,
        thread_adapter_type: type[ThreadAdapter] = ResponsesThreadAdapter,
    ):
        resolved_threading_mode = self.resolve_threading_mode(
            threading_mode=threading_mode,
            input_path=input_path,
        )

        if resolved_threading_mode == "auto" and thread_name_adapter is None:
            raise ValueError(
                "`llm_adapter` is required when `threading_mode` is 'auto'"
            )

        self.threading_mode = resolved_threading_mode
        self.input_path = resolved_threading_mode == "manual"
        self.thread_dir = thread_dir
        if thread_name_rules is not None and len(thread_name_rules) > 0:
            self.thread_name_rules = [*thread_name_rules]
        else:
            self.thread_name_rules = [*DEFAULT_THREAD_NAME_RULES]
        self._thread_name_adapter = thread_name_adapter
        self._thread_adapter_type = thread_adapter_type

        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            supports_tools=supports_tools,
            input_schema=input_schema,
            output_schema=output_schema,
            annotations=annotations,
            toolkits=toolkits,
        )

    def _sanitize_thread_name(self, *, value: str) -> str:
        normalized = value.strip().lower()
        if normalized.endswith(".thread"):
            normalized = normalized[: -len(".thread")]

        normalized = re.sub(r"[^a-z0-9]+", "-", normalized)
        normalized = re.sub(r"-{2,}", "-", normalized).strip("-")
        if normalized == "":
            normalized = "thread"
        return normalized[:64].strip("-") or "thread"

    def _fallback_thread_name(self, *, prompt: str) -> str:
        del prompt
        return f"thread-{datetime.now(timezone.utc).strftime('%Y%m%d-%H%M%S')}"

    def _thread_path_for_name(self, *, thread_name: str) -> str:
        return posixpath.join(self.thread_dir, f"{thread_name}.thread")

    async def _next_available_thread_path(
        self,
        *,
        context: TaskContext,
        base_path: str,
    ) -> str:
        try:
            exists = await context.room.storage.exists(path=base_path)
        except Exception:
            return base_path

        if not exists:
            return base_path

        thread_dir, filename = posixpath.split(base_path)
        if filename.endswith(".thread"):
            base_name = filename[: -len(".thread")]
        else:
            base_name = filename

        for index in range(2, 1000):
            candidate = posixpath.join(thread_dir, f"{base_name} {index}.thread")
            try:
                if not await context.room.storage.exists(path=candidate):
                    return candidate
            except Exception:
                return candidate

        return posixpath.join(thread_dir, f"{base_name}-{uuid.uuid4().hex[:8]}.thread")

    def _utc_now_iso(self) -> str:
        return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

    def _thread_list_index_path(self) -> str:
        return posixpath.join(self.thread_dir, "index.threadl")

    def _thread_list_entry_name_for_path(self, *, path: str) -> str:
        filename = posixpath.basename(path.strip())
        if filename.endswith(".thread"):
            filename = filename[: -len(".thread")]

        normalized = re.sub(r"[-_/]+", " ", filename)
        normalized = re.sub(r"\s+", " ", normalized).strip(" .-_")
        if normalized == "":
            return "Thread"
        if normalized == normalized.lower() or normalized == normalized.upper():
            normalized = normalized.title()
        return normalized[:64].strip() or "Thread"

    def _find_thread_list_entry(
        self,
        *,
        document: MeshDocument,
        path: str,
    ) -> Optional[Element]:
        for child in document.root.get_children():
            if not isinstance(child, Element):
                continue
            if child.tag_name != "thread":
                continue
            if child.get_attribute("path") == path:
                return child
        return None

    def _upsert_thread_list_entry(
        self,
        *,
        document: MeshDocument,
        path: str,
        name: Optional[str] = None,
        created_at: Optional[str] = None,
        modified_at: Optional[str] = None,
    ) -> None:
        now = self._utc_now_iso()
        provided_name = name.strip() if isinstance(name, str) else ""

        entry = self._find_thread_list_entry(document=document, path=path)
        if entry is None:
            resolved_name = (
                provided_name
                if provided_name != ""
                else self._thread_list_entry_name_for_path(path=path)
            )
            created_value = (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
            modified_value = (
                modified_at.strip()
                if isinstance(modified_at, str) and modified_at.strip() != ""
                else created_value
            )
            document.root.append_child(
                tag_name="thread",
                attributes={
                    "name": resolved_name,
                    "path": path,
                    "created_at": created_value,
                    "modified_at": modified_value,
                },
            )
            return

        if provided_name != "":
            entry.set_attribute("name", provided_name)
        else:
            existing_name = entry.get_attribute("name")
            if not isinstance(existing_name, str) or existing_name.strip() == "":
                entry.set_attribute(
                    "name",
                    self._thread_list_entry_name_for_path(path=path),
                )

        entry.set_attribute("path", path)

        existing_created_at = entry.get_attribute("created_at")
        created_value = (
            existing_created_at.strip()
            if isinstance(existing_created_at, str)
            and existing_created_at.strip() != ""
            else (
                created_at.strip()
                if isinstance(created_at, str) and created_at.strip() != ""
                else now
            )
        )
        entry.set_attribute("created_at", created_value)

        modified_value = (
            modified_at.strip()
            if isinstance(modified_at, str) and modified_at.strip() != ""
            else now
        )
        entry.set_attribute("modified_at", modified_value)

    async def record_thread_in_index(
        self,
        *,
        context: TaskContext,
        path: str,
    ) -> None:
        if self.threading_mode != "auto":
            return

        normalized_path = path.strip()
        if normalized_path == "":
            return

        index_path = self._thread_list_index_path()
        document = None
        try:
            document = await context.room.sync.open(
                path=index_path,
                schema=thread_list_schema,
            )
            now = self._utc_now_iso()
            self._upsert_thread_list_entry(
                document=document,
                path=normalized_path,
                created_at=now,
                modified_at=now,
            )
            state = document.get_state()
            if isinstance(state, bytes) and len(state) > 0:
                await context.room.sync.sync(
                    path=index_path,
                    data=base64.standard_b64encode(state),
                )
        except Exception as ex:
            logger.warning(
                "unable to update thread list document at %s",
                index_path,
                exc_info=ex,
            )
        finally:
            if document is not None:
                try:
                    await context.room.sync.close(path=index_path)
                except Exception as ex:
                    logger.warning(
                        "unable to close thread list document at %s",
                        index_path,
                        exc_info=ex,
                    )

    def ensure_local_member_on_thread(
        self,
        *,
        context: TaskContext,
        thread_adapter: ThreadAdapter,
    ) -> None:
        thread = thread_adapter.thread
        if thread is None:
            return

        members_elements = thread.root.get_children_by_tag_name("members")
        if len(members_elements) == 0:
            return

        members = members_elements[0]
        local_name = context.room.local_participant.get_attribute("name")
        if not isinstance(local_name, str):
            return

        normalized_local_name = local_name.strip()
        if normalized_local_name == "":
            return

        for child in members.get_children():
            if not isinstance(child, Element):
                continue
            if child.tag_name != "member":
                continue
            member_name = child.get_attribute("name")
            if isinstance(member_name, str) and member_name == normalized_local_name:
                return

        members.append_child(
            tag_name="member",
            attributes={"name": normalized_local_name},
        )

    async def _generate_thread_path(
        self,
        *,
        context: TaskContext,
        prompt: str,
        model: str,
    ) -> str:
        if self._thread_name_adapter is None:
            raise RuntimeError(
                "auto threading mode requires a configured llm adapter for thread naming"
            )

        cloned_context = context.session.copy()
        generated_name = self._fallback_thread_name(prompt=prompt)
        async with cloned_context:
            cloned_context.replace_rules(rules=self.thread_name_rules)
            cloned_context.append_user_message(prompt)

            try:
                response = await self._thread_name_adapter.next(
                    context=cloned_context,
                    room=context.room,
                    model=model,
                    on_behalf_of=context.on_behalf_of,
                    toolkits=[],
                    output_schema={
                        "type": "object",
                        "required": ["thread_name"],
                        "additionalProperties": False,
                        "properties": {
                            "thread_name": {
                                "type": "string",
                                "description": "2-6 word topic name for the task thread",
                            },
                        },
                    },
                )
                if isinstance(response, dict):
                    thread_name = response.get("thread_name")
                    if isinstance(thread_name, str):
                        generated_name = self._sanitize_thread_name(value=thread_name)
            except Exception as ex:
                logger.warning(
                    "unable to auto-generate thread name, using fallback", exc_info=ex
                )

        base_path = self._thread_path_for_name(thread_name=generated_name)
        return await self._next_available_thread_path(
            context=context,
            base_path=base_path,
        )

    async def resolve_thread_path(
        self,
        *,
        context: TaskContext,
        arguments: dict,
        prompt: str,
        model: str,
    ) -> str | None:
        if self.threading_mode == "none":
            return None

        if self.threading_mode == "manual":
            return self._selected_thread_path(arguments=arguments)

        return await self._generate_thread_path(
            context=context,
            prompt=prompt,
            model=model,
        )

    def _selected_thread_path(self, *, arguments: dict) -> str | None:
        if self.threading_mode == "none":
            return None

        path = arguments.get("path")
        if path is None:
            return None
        if not isinstance(path, str):
            raise ValueError("`path` must be a string or null")

        selected_path = path.strip()
        if selected_path == "":
            return None
        return selected_path

    def create_thread_adapter(
        self,
        *,
        context: TaskContext,
        arguments: dict,
        attachment: Optional[bytes] = None,
    ) -> ThreadAdapter | None:
        del attachment
        selected_path = self._selected_thread_path(arguments=arguments)
        if selected_path is None:
            return None

        return self._thread_adapter_type(
            room=context.room,
            path=selected_path,
        )



from __future__ import annotations

from email.message import EmailMessage
from typing import Any, Iterable, Iterator, Literal
import email.utils
import os
import re
import socket
import uuid

import mistune

type MessageRole = Literal["user", "agent"]

_DSN_STATUS_RE = re.compile(r"\b([245]\.\d{1,3}\.\d{1,3})\b")
_SMTP_DIAG_RE = re.compile(r"\b([245]\d\d)\b")


class SmtpConfiguration:
    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        port: int | None = None,
        hostname: str | None = None,
        local_hostname: str | None = None,
    ) -> None:
        if username is None:
            username = os.getenv("SMTP_USERNAME")

        if password is None:
            password = os.getenv("SMTP_PASSWORD")

        if port is None:
            port = int(os.getenv("SMTP_PORT", "587"))

        if hostname is None:
            hostname = os.getenv("SMTP_HOSTNAME")

        if local_hostname is None:
            local_hostname = os.getenv("SMTP_LOCAL_HOSTNAME")

        self.username = username
        self.password = password
        self.port = port
        self.hostname = hostname
        self.local_hostname = local_hostname

    def effective_local_hostname(self) -> str:
        candidates = [
            self.local_hostname,
            socket.getfqdn(),
            socket.gethostname(),
            os.getenv("HOSTNAME"),
            "localhost",
        ]
        for candidate in candidates:
            if candidate is None:
                continue
            normalized = candidate.strip()
            if normalized != "":
                return normalized

        return "localhost"


def _parse_addrs(values: Any) -> set[str]:
    if not values:
        return set()
    if isinstance(values, str):
        values = [values]
    return {
        addr.casefold() for _, addr in email.utils.getaddresses(list(values)) if addr
    }


def _first_addr(value: str | None) -> str | None:
    if not value:
        return None
    _, addr = email.utils.parseaddr(value)
    return addr or None


def _fmt_addr_list(addrs: Iterable[str]) -> str:
    return ", ".join(addrs)


def _get_first_header(msg: EmailMessage, name: str) -> str | None:
    value = msg.get(name)
    return str(value).strip() if value is not None else None


def _get_all_headers(msg: EmailMessage, name: str) -> list[str]:
    values = msg.get_all(name) or []
    return [str(v) for v in values]


def _clean_header_list(values: Any) -> list[str]:
    if not values:
        return []
    return [v for v in (str(x).strip() for x in values) if v]


def _parse_dsn_fields(msg: EmailMessage) -> dict[str, Any]:
    out: dict[str, Any] = {
        "is_dsn": False,
        "status": None,
        "action": None,
        "diagnostic_code": None,
        "final_recipient": None,
    }

    report_type = msg.get_param("report-type", header="Content-Type")
    if (
        msg.get_content_type().lower() == "multipart/report"
        and isinstance(report_type, str)
        and report_type.lower() == "delivery-status"
    ):
        out["is_dsn"] = True

    if msg.is_multipart():
        for part in msg.walk():
            if part.get_content_type().lower() != "message/delivery-status":
                continue

            out["is_dsn"] = True
            payload = part.get_payload()
            blocks = payload if isinstance(payload, list) else [payload]
            for block in blocks:
                if block is None:
                    continue
                action = block.get("Action")
                status = block.get("Status")
                diagnostic = block.get("Diagnostic-Code")
                final_recipient = block.get("Final-Recipient")

                if action and not out["action"]:
                    out["action"] = str(action)
                if status and not out["status"]:
                    out["status"] = str(status)
                if diagnostic and not out["diagnostic_code"]:
                    out["diagnostic_code"] = str(diagnostic)
                if final_recipient and not out["final_recipient"]:
                    out["final_recipient"] = str(final_recipient)
            break

    if not out["status"]:
        out["status"] = _get_first_header(msg, "Status")
    if not out["action"]:
        out["action"] = _get_first_header(msg, "Action")
    if not out["diagnostic_code"]:
        out["diagnostic_code"] = _get_first_header(msg, "Diagnostic-Code")
    if not out["final_recipient"]:
        out["final_recipient"] = _get_first_header(msg, "Final-Recipient")

    if out["status"]:
        match = _DSN_STATUS_RE.search(str(out["status"]))
        out["status"] = match.group(1) if match else str(out["status"]).strip()

    if out["diagnostic_code"]:
        diagnostic_match = _SMTP_DIAG_RE.search(str(out["diagnostic_code"]))
        if diagnostic_match is not None:
            out["diagnostic_code"] = diagnostic_match.group(1)

    return out


def _looks_like_error_or_autoreply(message: dict[str, Any]) -> tuple[bool, str]:
    meta = message.get("meta") or {}
    dsn = meta.get("dsn") or {}

    if dsn.get("is_dsn"):
        status = (dsn.get("status") or "").strip()
        diagnostic = (dsn.get("diagnostic_code") or "").strip()
        return True, f"DSN detected (status={status!r}, diagnostic={diagnostic!r})"

    content_type = (meta.get("content_type") or "").casefold()
    if content_type == "multipart/report":
        return True, "multipart/report detected"

    auto_submitted = (meta.get("auto_submitted") or "").casefold()
    if auto_submitted and auto_submitted != "no":
        return True, f"Auto-Submitted={meta.get('auto_submitted')!r}"

    precedence = (meta.get("precedence") or "").casefold()
    if precedence in {"bulk", "junk", "list", "auto_reply", "auto-reply"}:
        return True, f"Precedence={meta.get('precedence')!r}"

    return_path = (meta.get("return_path") or "").strip()
    if return_path == "<>":
        return True, "Return-Path is <>"

    from_header = message.get("from") or ""
    _, from_addr = email.utils.parseaddr(from_header)
    normalized_from = from_addr.casefold()
    if "mailer-daemon" in normalized_from or normalized_from.startswith("postmaster@"):
        return True, f"From looks like system sender ({from_addr or from_header})"

    subject = (message.get("subject") or "").casefold()
    for snippet in (
        "delivery status notification",
        "undelivered mail",
        "returned to sender",
        "mail delivery subsystem",
        "failure notice",
    ):
        if snippet in subject:
            return True, f"Subject looks like a bounce ({subject!r})"

    return False, ""


def message_to_json(*, message: EmailMessage, role: MessageRole) -> dict[str, Any]:
    body_part = message.get_body(("plain", "html"))
    if body_part is not None:
        body = body_part.get_content()
        body_content_type = body_part.get_content_type()
    else:
        body = message.get_content()
        body_content_type = message.get_content_type()

    message_id = message.get("Message-ID")
    if message_id is None:
        from_header = message.get("From", "")
        _, address = email.utils.parseaddr(from_header)
        domain = (address.split("@")[-1] if "@" in address else "local").lower()
        message_id = f"{uuid.uuid4()}@{domain}"

    to_list = _clean_header_list(_get_all_headers(message, "To"))
    cc_list = _clean_header_list(_get_all_headers(message, "Cc"))
    dsn = _parse_dsn_fields(message)

    return {
        "id": message_id,
        "in_reply_to": message.get("In-Reply-To"),
        "reply_to": message.get("Reply-To", message.get("From")),
        "references": message.get("References"),
        "from": message.get("From"),
        "to": to_list,
        "cc": cc_list,
        "subject": message.get("Subject"),
        "body": body,
        "attachments": [],
        "role": role,
        "correlation_id": message.get("Meshagent-Correlation-ID"),
        "meta": {
            "delivered_to": _get_first_header(message, "Delivered-To"),
            "return_path": _get_first_header(message, "Return-Path"),
            "to": to_list,
            "cc": cc_list,
            "auto_submitted": _get_first_header(message, "Auto-Submitted"),
            "precedence": _get_first_header(message, "Precedence"),
            "list_id": _get_first_header(message, "List-Id"),
            "content_type": message.get_content_type(),
            "is_multipart": message.is_multipart(),
            "body_content_type": body_content_type,
            "dsn": dsn,
        },
    }


def iter_message_attachments(message: EmailMessage) -> Iterator[tuple[str, bytes]]:
    for part in message.iter_attachments():
        file_name = part.get_filename() or "attachment.bin"
        data = part.get_content()
        if isinstance(data, str):
            yield file_name, data.encode(part.get_content_charset("utf-8"))
        else:
            yield file_name, data


def should_reply_to_message(
    *,
    message: dict[str, Any],
    email_address: str,
    whitelist: list[str] | None = None,
) -> tuple[bool, str | None]:
    my_addr = email_address.casefold()
    to_addrs = _parse_addrs(message.get("to"))
    cc_addrs = _parse_addrs(message.get("cc"))
    delivered_to = ((message.get("meta") or {}).get("delivered_to") or "").casefold()

    addressed = (
        (my_addr in to_addrs) or (my_addr in cc_addrs) or (delivered_to == my_addr)
    )
    if not addressed:
        return (
            False,
            "message is not addressed to this mailbox",
        )

    is_error, reason = _looks_like_error_or_autoreply(message)
    if is_error:
        return False, reason

    if whitelist is not None:
        from_header = message.get("from") or ""
        _, addr = email.utils.parseaddr(from_header)
        if addr.casefold() not in {entry.casefold() for entry in whitelist}:
            return False, f"{from_header} not found in whitelist"

    return True, None


def render_mail_markdown(*, body: str) -> str:
    markdown = mistune.create_markdown(plugins=["table"])
    return markdown(body)


def create_email_message(
    *,
    to_address: str,
    from_address: str,
    subject: str,
    body: str,
    correlation_id: str | None = None,
) -> EmailMessage:
    _, address = email.utils.parseaddr(from_address)
    domain = (address.split("@")[-1] if "@" in address else "local").lower()
    message_id = f"<{uuid.uuid4()}@{domain}>"

    message = EmailMessage()
    message["Message-ID"] = message_id
    message["Subject"] = subject
    message["From"] = from_address
    message["To"] = to_address
    if correlation_id is not None:
        message["Meshagent-Correlation-ID"] = correlation_id

    message.set_content(body)
    message.add_alternative(render_mail_markdown(body=body), subtype="html")
    return message


def create_reply_email_message(
    *,
    message: dict[str, Any],
    from_address: str,
    body: str,
    email_address: str,
    reply_all: bool = False,
) -> EmailMessage:
    subject = str(message.get("subject") or "")
    if not subject.lower().startswith("re:"):
        subject = "RE: " + subject

    _, address = email.utils.parseaddr(from_address)
    domain = (address.split("@")[-1] if "@" in address else "local").lower()
    message_id = f"<{uuid.uuid4()}@{domain}>"

    reply_to_header = str(message.get("reply_to") or message.get("from") or "")
    reply_to_addr = _first_addr(reply_to_header)

    my_addr = email_address.casefold()
    original_to = _parse_addrs(message.get("to"))
    original_cc = _parse_addrs(message.get("cc"))

    to_addrs: list[str] = []
    cc_addrs: list[str] = []

    if reply_to_addr:
        to_addrs = [reply_to_addr]

    if reply_all:
        sender_casefold = (reply_to_addr or "").casefold()
        current_to = {item.casefold() for item in to_addrs}
        for address_value in original_to | original_cc:
            normalized = address_value.casefold()
            if normalized == my_addr:
                continue
            if sender_casefold != "" and normalized == sender_casefold:
                continue
            if normalized in current_to:
                continue
            cc_addrs.append(address_value)

    reply_message = EmailMessage()
    reply_message["Message-ID"] = message_id
    reply_message["Subject"] = subject
    reply_message["From"] = from_address
    if to_addrs:
        reply_message["To"] = _fmt_addr_list(to_addrs)
    else:
        reply_message["To"] = reply_to_header
    if len(cc_addrs) > 0:
        reply_message["Cc"] = _fmt_addr_list(cc_addrs)

    parent_message_id = message.get("id")
    if isinstance(parent_message_id, str) and parent_message_id.strip() != "":
        reply_message["In-Reply-To"] = parent_message_id

    references = message.get("references")
    if isinstance(references, str) and references.strip() != "":
        reply_message["References"] = references

    correlation_id = message.get("correlation_id")
    if isinstance(correlation_id, str) and correlation_id.strip() != "":
        reply_message["Meshagent-Correlation-ID"] = correlation_id

    reply_message.set_content(body)
    reply_message.add_alternative(render_mail_markdown(body=body), subtype="html")
    return reply_message



from __future__ import annotations

import asyncio
import contextlib
from datetime import datetime, timezone
import json
import logging
from pathlib import PurePosixPath
import re
from typing import Any
from urllib.parse import urlparse
import uuid

from meshagent.api import Participant, RoomClient
from pydantic import TypeAdapter, ValidationError

from .adapter import LLMAdapter
from .messages import (
    AgentFileContent,
    AgentInputContent,
    AgentTextContent,
    TurnStart,
)
from .threaded_channel import ThreadedChannel

logger = logging.getLogger("queue-channel")

_agent_input_content_list_adapter = TypeAdapter(list[AgentInputContent])
_thread_id_template_pattern = re.compile(r"\{([^{}]+)\}")


class QueueChannel(ThreadedChannel):
    def __init__(
        self,
        *,
        room: RoomClient,
        queue_name: str,
        threading_mode: str | None = None,
        thread_dir: str | None = None,
        llm_adapter: LLMAdapter | None = None,
    ) -> None:
        super().__init__(
            room=room,
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            llm_adapter=llm_adapter,
        )
        normalized_queue_name = queue_name.strip()
        if normalized_queue_name == "":
            raise ValueError("queue_name must not be empty")

        self._queue_name = normalized_queue_name
        self._receive_task: asyncio.Task[None] | None = None

    def _default_thread_dir_fallback_name(self) -> str:
        return "queue"

    def handles(self, message) -> bool:
        del message
        return False

    async def on_start(self) -> None:
        await self.publish_thread_attributes()
        await self.open_thread_list_document()
        self._receive_task = asyncio.create_task(self._receive_loop())

    async def on_stop(self) -> None:
        receive_task = self._receive_task
        self._receive_task = None
        if receive_task is not None:
            receive_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await receive_task
        await self.close_thread_list_document()

    async def _receive_loop(self) -> None:
        while not self._stop.is_set():
            try:
                queued_message = await self._room.queues.receive(
                    name=self._queue_name,
                    create=True,
                    wait=True,
                )
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.exception(
                    "queue receive failed for queue %s",
                    self._queue_name,
                )
                await asyncio.sleep(1)
                continue

            if queued_message is None:
                continue

            try:
                logger.info(f"processing message from queue {self._queue_name}")
                payload = self._payload_from_queue_message(message=queued_message)
                content = await self._content_from_payload(payload=payload)
                thread_text = self._thread_text_from_content(
                    payload=payload,
                    content=content,
                )
                self.emit(
                    sender=self._sender_from_payload(payload=payload),
                    payload=TurnStart(
                        type="meshagent.agent.turn.start",
                        thread_id=await self._thread_id_from_payload(
                            payload=payload,
                            thread_text=thread_text,
                        ),
                        content=content,
                        toolkits=self._toolkits_from_payload(payload=payload),
                        model=self._model_from_payload(payload=payload),
                        instructions=self._instructions_from_payload(payload=payload),
                    ),
                )
            except asyncio.CancelledError:
                raise
            except Exception:
                logger.exception(
                    "queue channel failed to process a message from %s",
                    self._queue_name,
                )

    def _payload_from_queue_message(self, *, message: Any) -> dict[str, Any]:
        if isinstance(message, dict):
            return dict(message)

        if isinstance(message, str):
            return {"prompt": message}

        return {
            "prompt": json.dumps(
                message,
                ensure_ascii=False,
                default=str,
            )
        }

    async def _content_from_payload(
        self,
        *,
        payload: dict[str, Any],
    ) -> list[AgentInputContent]:
        source, content = self._content_items_from_payload(payload=payload)
        legacy_prompt_file_text = await self._legacy_prompt_file_text_from_payload(
            payload=payload
        )
        if legacy_prompt_file_text is not None:
            content = [
                AgentTextContent(type="text", text=legacy_prompt_file_text),
                *content,
            ]

        if source == "prompt":
            content = await self._resolve_prompt_file_content_items(content=content)

        if len(content) > 0:
            return content

        return [
            AgentTextContent(
                type="text",
                text=json.dumps(payload, ensure_ascii=False, default=str),
            )
        ]

    def _content_items_from_payload(
        self,
        *,
        payload: dict[str, Any],
    ) -> tuple[str, list[AgentInputContent]]:
        raw_content = payload.get("content")
        if raw_content is not None:
            return (
                "content",
                self._validate_content_items(raw_content=raw_content, key="content"),
            )

        prompt = payload.get("prompt")
        if isinstance(prompt, str):
            if prompt.strip() != "":
                return ("prompt", [AgentTextContent(type="text", text=prompt)])
            if self._has_legacy_prompt_file(payload=payload):
                return ("prompt", [])
        elif isinstance(prompt, (dict, list)):
            return (
                "prompt",
                self._validate_content_items(raw_content=prompt, key="prompt"),
            )

        if self._has_legacy_prompt_file(payload=payload):
            return ("prompt", [])

        logger.warning(
            "prompt property not found on queue message from %s, inserting whole message into context",
            self._queue_name,
        )
        return (
            "fallback",
            [
                AgentTextContent(
                    type="text",
                    text=json.dumps(payload, ensure_ascii=False, default=str),
                )
            ],
        )

    @staticmethod
    def _validate_content_items(
        *,
        raw_content: dict[str, Any] | list[Any],
        key: str,
    ) -> list[AgentInputContent]:
        content_input: list[Any]
        if isinstance(raw_content, dict):
            content_input = [raw_content]
        elif isinstance(raw_content, list):
            content_input = raw_content
        else:
            raise ValueError(f"queue message {key} must be AgentInputContent JSON")

        try:
            return _agent_input_content_list_adapter.validate_python(content_input)
        except ValidationError as exc:
            raise ValueError(
                f"queue message {key} must be AgentInputContent JSON"
            ) from exc

    async def _legacy_prompt_file_text_from_payload(
        self,
        *,
        payload: dict[str, Any],
    ) -> str | None:
        prompt_file = payload.get("prompt_file")
        if not isinstance(prompt_file, str):
            return None

        normalized_prompt_file = prompt_file.strip()
        if normalized_prompt_file == "":
            return None

        try:
            storage_path = self._normalize_room_storage_path(
                path=normalized_prompt_file
            )
        except ValueError:
            logger.warning(
                "invalid prompt_file %s on queue %s",
                normalized_prompt_file,
                self._queue_name,
            )
            return None

        return await self._resolve_prompt_file_content(path=storage_path)

    async def _resolve_prompt_file_content_items(
        self,
        *,
        content: list[AgentInputContent],
    ) -> list[AgentInputContent]:
        resolved_content: list[AgentInputContent] = []
        for item in content:
            resolved_item = await self._resolve_prompt_content_item(item=item)
            if resolved_item is None:
                continue

            resolved_content.append(resolved_item)

        return resolved_content

    async def _resolve_prompt_content_item(
        self,
        *,
        item: AgentInputContent,
    ) -> AgentInputContent | None:
        if not isinstance(item, AgentFileContent):
            return item

        try:
            storage_path = self._room_storage_path_from_url(url=item.url)
        except ValueError:
            logger.warning(
                "invalid room prompt file url %s on queue %s",
                item.url,
                self._queue_name,
            )
            return None

        if storage_path is None:
            return item

        resolved_prompt_text = await self._resolve_prompt_file_content(
            path=storage_path
        )
        if resolved_prompt_text is None:
            return None

        return AgentTextContent(
            type="text",
            text=resolved_prompt_text,
        )

    async def _resolve_prompt_file_content(
        self,
        *,
        path: str,
    ) -> str | None:
        if not await self._room.storage.exists(path=path):
            logger.warning(
                "prompt_file %s not found in room storage for queue %s",
                path,
                self._queue_name,
            )
            return None

        prompt_file_content = await self._room.storage.download(path=path)
        try:
            return prompt_file_content.data.decode("utf-8")
        except UnicodeDecodeError:
            logger.warning(
                "prompt_file %s is not valid utf-8 for queue %s",
                path,
                self._queue_name,
            )
            return None

    @staticmethod
    def _normalize_room_storage_path(*, path: str) -> str:
        normalized = PurePosixPath("/" + path.strip().lstrip("/")).as_posix().strip("/")
        if normalized == "":
            raise ValueError("room storage path must reference a non-root path")

        if any(part in {".", ".."} for part in PurePosixPath(normalized).parts):
            raise ValueError("room storage path cannot contain '.' or '..' segments")

        return normalized

    @classmethod
    def _room_storage_path_from_url(cls, *, url: str) -> str | None:
        parsed_url = urlparse(url)
        if parsed_url.scheme == "":
            return None
        if parsed_url.scheme != "room":
            return None

        raw_path = f"{parsed_url.netloc}{parsed_url.path}"
        return cls._normalize_room_storage_path(path=raw_path)

    def _has_legacy_prompt_file(self, *, payload: dict[str, Any]) -> bool:
        prompt_file = payload.get("prompt_file")
        return isinstance(prompt_file, str) and prompt_file.strip() != ""

    def _thread_text_from_content(
        self,
        *,
        payload: dict[str, Any],
        content: list[AgentInputContent],
    ) -> str:
        text_parts = [
            item.text
            for item in content
            if isinstance(item, AgentTextContent) and item.text.strip() != ""
        ]
        if len(text_parts) > 0:
            return "\n\n".join(text_parts)

        return json.dumps(payload, ensure_ascii=False, default=str)

    async def _thread_id_from_payload(
        self,
        *,
        payload: dict[str, Any],
        thread_text: str,
    ) -> str:
        for key in ("thread_id", "path"):
            value = payload.get(key)
            if isinstance(value, str) and value.strip() != "":
                return self._expand_thread_id_template(value=value.strip())

        path, _ = await self.new_thread(
            message_text=thread_text,
        )
        return path

    def _expand_thread_id_template(self, *, value: str) -> str:
        now = self._now()
        year = now.strftime("%Y")
        short_year = now.strftime("%y")
        month = now.strftime("%m")
        day = now.strftime("%d")
        hour = now.strftime("%H")
        minute = now.strftime("%M")
        second = now.strftime("%S")

        # Keep the legacy short month/minute forms exact-case to avoid the
        # MM/mm ambiguity, and offer unambiguous aliases that are matched
        # case-insensitively.
        legacy_exact_replacements = {
            "YYYY": year,
            "YY": short_year,
            "MM": month,
            "DD": day,
            "HH": hour,
            "mm": minute,
            "SS": second,
            "ss": second,
        }
        case_insensitive_replacements = {
            "YYYY": year,
            "YEAR": year,
            "YY": short_year,
            "YR": short_year,
            "DD": day,
            "DAY": day,
            "HH": hour,
            "HOUR": hour,
            "SS": second,
            "SECOND": second,
            "SEC": second,
            "MONTH": month,
            "MON": month,
            "MINUTE": minute,
            "MIN": minute,
        }

        def replace_match(match: re.Match[str]) -> str:
            token = match.group(1)
            exact_replacement = legacy_exact_replacements.get(token)
            if exact_replacement is not None:
                return exact_replacement

            return case_insensitive_replacements.get(
                token.upper(),
                match.group(0),
            )

        return _thread_id_template_pattern.sub(replace_match, value)

    def _now(self) -> datetime:
        return datetime.now(timezone.utc)

    def _sender_from_payload(self, *, payload: dict[str, Any]) -> Participant | None:
        sender_name = payload.get("sender_name")
        if not isinstance(sender_name, str) or sender_name.strip() == "":
            sender_name = payload.get("from")
            if not isinstance(sender_name, str) or sender_name.strip() == "":
                return None

        normalized_sender = sender_name.strip()
        return Participant(
            id=f"queue:{self._queue_name}:{uuid.uuid5(uuid.NAMESPACE_URL, normalized_sender)}",
            attributes={"name": normalized_sender},
        )

    @staticmethod
    def _toolkits_from_payload(
        *, payload: dict[str, Any]
    ) -> list[dict[str, Any]] | None:
        raw_toolkits = payload.get("tools")
        if not isinstance(raw_toolkits, list):
            return None

        toolkits = [toolkit for toolkit in raw_toolkits if isinstance(toolkit, dict)]
        if len(toolkits) == 0:
            return None
        return toolkits

    @staticmethod
    def _model_from_payload(*, payload: dict[str, Any]) -> str | None:
        model = payload.get("model")
        if not isinstance(model, str) or model.strip() == "":
            return None
        return model

    @staticmethod
    def _instructions_from_payload(*, payload: dict[str, Any]) -> str | None:
        instructions = payload.get("instructions")
        if not isinstance(instructions, str) or instructions.strip() == "":
            return None
        return instructions



from meshagent.api.schema import MeshSchema, ElementType, ChildProperty, ValueProperty


thread_list_schema = MeshSchema(
    root_tag_name="thread_list",
    elements=[
        ElementType(
            tag_name="thread_list",
            description="an index of chat threads",
            properties=[
                ChildProperty(
                    name="threads",
                    description="threads known to this chatbot",
                    child_tag_names=["thread"],
                )
            ],
        ),
        ElementType(
            tag_name="thread",
            description="a thread entry in the index",
            properties=[
                ValueProperty(
                    name="name",
                    description="a readable name for the thread",
                    type="string",
                ),
                ValueProperty(
                    name="path",
                    description="the sync path to the thread document",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the date the thread was created in ISO format",
                    type="string",
                ),
                ValueProperty(
                    name="modified_at",
                    description="the date the thread was last modified in ISO format",
                    type="string",
                ),
            ],
        ),
    ],
)

thread_schema = MeshSchema(
    root_tag_name="thread",
    elements=[
        ElementType(
            tag_name="thread",
            description="a thread of messages",
            properties=[
                ValueProperty(
                    name="name", description="name of the tread", type="string"
                ),
                ChildProperty(
                    name="properties",
                    description="the messages in the thread",
                    ordered=True,
                    child_tag_names=["members", "messages"],
                ),
            ],
        ),
        ElementType(
            tag_name="members",
            description="the members of this thread",
            properties=[
                ChildProperty(
                    name="items",
                    child_tag_names=["member"],
                    description="the messages in this thread",
                )
            ],
        ),
        ElementType(
            tag_name="messages",
            description="the messages of this thread",
            properties=[
                ValueProperty(
                    name="external_thread_id",
                    description="optional external backend thread id used for resume",
                    type="string",
                ),
                ChildProperty(
                    name="items",
                    child_tag_names=[
                        "message",
                        "exec",
                        "ui",
                        "reasoning",
                        "event",
                    ],
                    description="the messages in this thread",
                ),
            ],
        ),
        ElementType(
            tag_name="member",
            description="a member of this thread",
            properties=[
                ValueProperty(
                    name="name", description="the name of the member", type="string"
                ),
            ],
        ),
        ElementType(
            tag_name="file",
            description="a file attachment",
            properties=[
                ValueProperty(
                    name="path",
                    description="the path of the file in the room",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="image",
            description="an image attachment stored in the images database",
            properties=[
                ValueProperty(
                    name="status",
                    description="the image generation status (for example generating/completed/failed)",
                    type="string",
                ),
                ValueProperty(
                    name="status_detail",
                    description="optional human-readable status text for image generation",
                    type="string",
                ),
                ValueProperty(
                    name="width",
                    description="the pixel width of the generated image when known",
                    type="number",
                ),
                ValueProperty(
                    name="height",
                    description="the pixel height of the generated image when known",
                    type="number",
                ),
                ValueProperty(
                    name="id",
                    description="the image id in the room images table",
                    type="string",
                ),
                ValueProperty(
                    name="mime_type",
                    description="the image mime type",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the date that the image was generated in ISO format",
                    type="string",
                ),
                ValueProperty(
                    name="created_by",
                    description="the participant who generated the image",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="reasoning",
            description="a reasoning trace",
            properties=[
                ValueProperty(
                    name="turn_id",
                    description="the agent turn id associated with this reasoning trace",
                    type="string",
                ),
                ValueProperty(
                    name="summary",
                    description="a summary of the reasoning",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the time that the reasoning started",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="exec",
            description="a command execution",
            properties=[
                ValueProperty(
                    name="turn_id",
                    description="the agent turn id associated with this command execution",
                    type="string",
                ),
                ValueProperty(
                    name="command",
                    description="a command that was executed",
                    type="string",
                ),
                ValueProperty(
                    name="result",
                    description="the result of the command",
                    type="string",
                ),
                ValueProperty(
                    name="outcome",
                    description="the outcome of the command",
                    type="string",
                ),
                ValueProperty(
                    name="stderr",
                    description="the standard error text of the command",
                    type="string",
                ),
                ValueProperty(
                    name="stdout",
                    description="the standard output text of the command",
                    type="string",
                ),
                ValueProperty(
                    name="exit_code",
                    description="the exit code of the command",
                    type="number",
                ),
                ValueProperty(
                    name="pwd",
                    description="the working directory the command was executed in",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the time that the reasoning started",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="ui",
            description="custom user interface data",
            properties=[
                ValueProperty(
                    name="turn_id",
                    description="the agent turn id associated with this UI payload",
                    type="string",
                ),
                ValueProperty(
                    name="data",
                    description="raw data to be renderered, usually in JSON format",
                    type="string",
                ),
                ValueProperty(
                    name="renderer",
                    description="the renderer to use",
                    type="string",
                ),
                ValueProperty(
                    name="widget",
                    description="the type of widget",
                    type="string",
                ),
                ValueProperty(
                    name="width",
                    description="the width of widget",
                    type="number",
                ),
                ValueProperty(
                    name="height",
                    description="the height of widget",
                    type="number",
                ),
                ValueProperty(
                    name="created_at",
                    description="the time that the reasoning started",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="event",
            description="a structured event emitted by an agent backend",
            properties=[
                ValueProperty(
                    name="turn_id",
                    description="the agent turn id associated with this event",
                    type="string",
                ),
                ValueProperty(
                    name="id",
                    description="a unique id for this event",
                    type="string",
                ),
                ValueProperty(
                    name="source",
                    description="backend source, for example codex/openai/anthropic/chatkit",
                    type="string",
                ),
                ValueProperty(
                    name="name",
                    description="normalized event name",
                    type="string",
                ),
                ValueProperty(
                    name="kind",
                    description="high-level event category, for example turn/item/plan/diff/exec",
                    type="string",
                ),
                ValueProperty(
                    name="state",
                    description="lifecycle state, for example queued/in_progress/completed/failed/cancelled/info",
                    type="string",
                ),
                ValueProperty(
                    name="method",
                    description="source event or rpc method",
                    type="string",
                ),
                ValueProperty(
                    name="item_id",
                    description="item id for this event, when available",
                    type="string",
                ),
                ValueProperty(
                    name="item_type",
                    description="item type for this event, when available",
                    type="string",
                ),
                ValueProperty(
                    name="path",
                    description="optional path associated with the event, typically a file path",
                    type="string",
                ),
                ValueProperty(
                    name="summary",
                    description="short human readable summary",
                    type="string",
                ),
                ValueProperty(
                    name="headline",
                    description="primary event headline for compact rendering",
                    type="string",
                ),
                ValueProperty(
                    name="details",
                    description="optional newline-delimited detail lines for structured rendering",
                    type="string",
                ),
                ValueProperty(
                    name="preview",
                    description="optional structured preview body for compact rendering, for example file contents",
                    type="string",
                ),
                ChildProperty(
                    name="logs",
                    description="bounded rolling logs for this event",
                    child_tag_names=["log"],
                ),
                ValueProperty(
                    name="data",
                    description="serialized backend payload",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the date this event was emitted in ISO format",
                    type="string",
                ),
                ValueProperty(
                    name="updated_at",
                    description="the date this event was last updated in ISO format",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="log",
            description="a single stdout or stderr log line emitted by an event",
            properties=[
                ValueProperty(
                    name="source",
                    description="the source stream for this log line",
                    type="string",
                ),
                ValueProperty(
                    name="text",
                    description="the log line text",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the date this log line was recorded in ISO format",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="reaction",
            description="a unicode reaction from a user on a message or specific attachment",
            properties=[
                ValueProperty(
                    name="user_name",
                    description="the display name of the user who reacted",
                    type="string",
                ),
                ValueProperty(
                    name="value",
                    description="the reaction as a single unicode character",
                    type="string",
                ),
                ValueProperty(
                    name="target",
                    description="reaction target type (message or attachment)",
                    type="string",
                ),
                ValueProperty(
                    name="attachment_ref",
                    description="the attachment element id when target is attachment",
                    type="string",
                ),
                ValueProperty(
                    name="created_at",
                    description="the date that the reaction was added in ISO format",
                    type="string",
                ),
            ],
        ),
        ElementType(
            tag_name="message",
            description="a message sent in the conversation",
            properties=[
                ValueProperty(
                    name="id", description="the id of the message", type="string"
                ),
                ValueProperty(
                    name="turn_id",
                    description="the agent turn id associated with this message",
                    type="string",
                ),
                ValueProperty(
                    name="text", description="the text of the message", type="string"
                ),
                ValueProperty(
                    name="created_at",
                    description="the date that the message was sent in ISO format",
                    type="string",
                ),
                ValueProperty(
                    name="author_name",
                    description="the name of the author of the post",
                    type="string",
                ),
                ValueProperty(
                    name="role",
                    description="the role of the message author, such as user, agent, or tool",
                    type="string",
                ),
                ValueProperty(
                    name="author_ref",
                    description="a reference to author identity in another system",
                    type="string",
                ),
                ChildProperty(
                    name="attachments",
                    child_tag_names=["file", "image", "reaction"],
                    description="a list of message attachments",
                ),
            ],
        ),
    ],
)



from typing import Optional

from jsonschema import validate, ValidationError
from meshagent.api.schema_util import prompt_schema, merge
from meshagent.api import Requirement
from meshagent.tools import Toolkit, make_toolkits, ToolkitBuilder
from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.completions_thread_adapter import CompletionsThreadAdapter
from meshagent.agents.task_runner import TaskContext
from meshagent.agents.responses_thread_adapter import ResponsesThreadAdapter
from meshagent.agents.threaded_task_runner import ThreadedTaskRunner, ThreadingMode
from meshagent.agents.toolkit_schema import build_tools_property_schema
from meshagent.openai.tools.completions_adapter import OpenAICompletionsAdapter

import tarfile
import io
import mimetypes

import logging

logger = logging.getLogger("llm-runner")

ThreadAdapter = ResponsesThreadAdapter
CompletionsAdapterThreadAdapter = CompletionsThreadAdapter


class LLMTaskRunner(ThreadedTaskRunner):
    """
    A Task Runner that uses an LLM execution loop until the task is complete.
    """

    def __init__(
        self,
        *,
        llm_adapter: LLMAdapter,
        title: Optional[str] = None,
        description: Optional[str] = None,
        toolkits: Optional[list[Toolkit]] = None,
        requires: Optional[list[Requirement]] = None,
        supports_tools: bool = True,
        input_prompt: bool = True,
        input_path: bool = False,
        threading_mode: Optional[ThreadingMode] = None,
        thread_dir: str = ".threads",
        thread_name_rules: Optional[list[str]] = None,
        input_schema: Optional[dict] = None,
        output_schema: Optional[dict] = None,
        allow_model_selection: bool = True,
        rules: Optional[list[str]] = None,
        annotations: Optional[list[str]] = None,
        client_rules: Optional[dict[str, list[str]]] = None,
    ):
        self.allow_model_selection = allow_model_selection
        resolved_threading_mode = self.resolve_threading_mode(
            threading_mode=threading_mode,
            input_path=input_path,
        )

        if input_schema is None:
            if input_prompt:
                input_schema = prompt_schema(
                    description="use a prompt to generate content"
                )

                if allow_model_selection:
                    input_schema = merge(
                        schema=input_schema,
                        additional_properties={
                            "model": {"type": ["string", "null"]},
                        },
                    )

                input_schema = self.with_manual_thread_path_schema(
                    input_schema=input_schema,
                    threading_mode=resolved_threading_mode,
                )

                tools_schema, defs = build_tools_property_schema(
                    toolkit_builders=self.get_toolkit_builders()
                )
                if tools_schema is not None:
                    input_schema = merge(
                        schema=input_schema,
                        additional_properties={
                            "tools": tools_schema,
                        },
                    )

                    if len(defs) > 0:
                        if input_schema.get("$defs") is None:
                            input_schema["$defs"] = {}

                        for key, value in defs.items():
                            input_schema["$defs"][key] = value

            else:
                input_schema = {
                    "type": "object",
                    "additionalProperties": False,
                    "required": [],
                    "properties": {},
                }

        static_toolkits = list(toolkits or [])
        thread_adapter_type = ThreadAdapter
        if isinstance(llm_adapter, OpenAICompletionsAdapter):
            thread_adapter_type = CompletionsAdapterThreadAdapter

        super().__init__(
            title=title,
            description=description,
            input_schema=input_schema,
            output_schema=output_schema,
            requires=requires,
            supports_tools=supports_tools,
            annotations=annotations,
            toolkits=static_toolkits,
            input_path=input_path,
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            thread_name_rules=thread_name_rules,
            thread_name_adapter=llm_adapter,
            thread_adapter_type=thread_adapter_type,
        )

        self._extra_rules = rules or []
        self._llm_adapter = llm_adapter
        self.toolkits = static_toolkits
        self._client_rules = client_rules

    async def init_session(self):
        chat = self._llm_adapter.create_session()
        return chat

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_context_toolkits(self, *, context: TaskContext) -> list[Toolkit]:
        return []

    async def get_rules(self, *, context: TaskContext):
        rules = [*self._extra_rules]

        participant = context.caller
        client = participant.get_attribute("client")

        if self._client_rules is not None and client is not None:
            cr = self._client_rules.get(client)
            if cr is not None:
                rules.extend(cr)

        return rules

    async def is_done(self, *, context: TaskContext):
        return context.session.turn_count > 0

    async def ask(
        self,
        *,
        context: TaskContext,
        arguments: dict,
        attachment: Optional[bytes] = None,
    ):
        prompt = arguments.get("prompt")
        if prompt is None:
            raise ValueError("`prompt` is required")

        message_tools = arguments.get("tools")
        if self.allow_model_selection:
            model = arguments.get("model", self._llm_adapter.default_model())
        else:
            model = self._llm_adapter.default_model()

        thread_path = await self.resolve_thread_path(
            context=context,
            arguments=arguments,
            prompt=prompt,
            model=model,
        )
        adapter_arguments = arguments
        if thread_path is not None:
            adapter_arguments = {
                **arguments,
                "path": thread_path,
            }
            await self.record_thread_in_index(
                context=context,
                path=thread_path,
            )

        thread_adapter = self.create_thread_adapter(
            context=context,
            arguments=adapter_arguments,
            attachment=attachment,
        )

        if thread_adapter is not None:
            await thread_adapter.start()
            self.ensure_local_member_on_thread(
                context=context,
                thread_adapter=thread_adapter,
            )
            thread_adapter.append_messages(context=context.session)
            thread_adapter.write_text_message(text=prompt, participant=context.caller)

        try:
            context.session.append_rules(await self.get_rules(context=context))

            context.session.append_user_message(prompt)

            if attachment is not None:
                buf = io.BytesIO(attachment)
                with tarfile.open(fileobj=buf, mode="r:*") as tar:
                    for member in tar.getmembers():
                        if member.isfile():
                            mime_type, encoding = mimetypes.guess_type(member.name)
                            del encoding
                            f = tar.extractfile(member)
                            if f is None:
                                continue
                            content = f.read()

                            normalized_mime_type = (
                                mime_type or "application/octet-stream"
                            )
                            if (
                                normalized_mime_type.startswith("image/")
                                and context.session.supports_images
                            ):
                                context.session.append_image_message(
                                    data=content, mime_type=normalized_mime_type
                                )
                            elif context.session.supports_files:
                                context.session.append_file_message(
                                    filename=member.name,
                                    data=content,
                                    mime_type=normalized_mime_type,
                                )
                            else:
                                context.session.append_user_message(
                                    f"the user attached a file named '{member.name}' with mime type '{normalized_mime_type}'"
                                )

            combined_toolkits: list[Toolkit] = [
                *self.toolkits,
                *context.toolkits,
                *await self.get_context_toolkits(context=context),
                *await self.get_required_toolkits(context=context),
            ]

            if message_tools is not None and len(message_tools) > 0:
                combined_toolkits.extend(
                    await make_toolkits(
                        room=self.room,
                        model=model,
                        providers=self.get_toolkit_builders(),
                        tools=message_tools,
                    )
                )

            def push(event: dict):
                if thread_adapter is not None:
                    thread_adapter.push(event=event)

            resp = None
            while not await self.is_done(context=context):
                try:
                    resp = await self._llm_adapter.next(
                        context=context.session,
                        room=context.room,
                        toolkits=combined_toolkits,
                        output_schema=self.output_schema,
                        event_handler=push,
                    )

                except Exception as ex:
                    logger.error("unexpected error during task execution", exc_info=ex)

            if resp is None:
                raise RuntimeError("LLM adapter did not return a response")

            # Validate the LLM output against the declared output schema if one was provided
            if self.output_schema:
                try:
                    validate(instance=resp, schema=self.output_schema)
                except ValidationError as exc:
                    raise RuntimeError("LLM output failed schema validation") from exc

            return resp

        finally:
            if thread_adapter is not None:
                await thread_adapter.stop()

    def create_thread_adapter(
        self,
        *,
        context: TaskContext,
        arguments: dict,
        attachment: Optional[bytes] = None,
    ):
        del attachment
        selected_path = self._selected_thread_path(arguments=arguments)
        if selected_path is None:
            return None

        if issubclass(self._thread_adapter_type, ResponsesThreadAdapter):
            return self._thread_adapter_type(
                room=context.room,
                path=selected_path,
            )

        return self._thread_adapter_type(
            room=context.room,
            path=selected_path,
        )



import asyncio
import base64
import json
import logging
import re
import shlex
import uuid
from datetime import datetime, timezone
from typing import Any, Optional

from meshagent.api import Element, RoomException

from .images_database import ImagesDatabase
from .image_captioner import ImageCaptioner
from .thread_adapter import ThreadAdapter, tracer

logger = logging.getLogger("thread_adapter")

_ACTIVE_STATES = {"queued", "in_progress", "running", "pending", "searching"}
_TERMINAL_STATES = {"completed", "failed", "cancelled"}
_RESPONSE_NOISE_TYPES = {
    "response.content_part.added",
    "response.content_part.done",
    "response.output_text.delta",
    "response.output_text.done",
    "response.refusal.delta",
    "response.refusal.done",
    "response.reasoning_text.delta",
    "response.reasoning_text.done",
    "response.reasoning_summary_text.delta",
    "response.reasoning_summary_text.done",
    "response.reasoning_summary_part.added",
    "response.reasoning_summary_part.done",
    "response.function_call_arguments.delta",
    "response.function_call_arguments.done",
    "response.mcp_call_arguments.delta",
    "response.mcp_call_arguments.done",
    "response.custom_tool_call_input.delta",
    "response.custom_tool_call_input.done",
    "response.image_generation_call.partial_image",
}
_RESPONSE_TURN_EVENTS = {
    "response.created",
    "response.in_progress",
    "response.completed",
    "response.failed",
    "response.cancelled",
    "response.canceled",
    "response.queued",
}
_SUPPORTED_EVENT_KINDS = {
    "exec",
    "tool",
    "web",
    "search",
    "diff",
    "image",
    "approval",
    "collab",
    "plan",
}
_IMAGE_DB_SAVE_TIMEOUT_SECONDS = 20.0
_IMAGE_STAGE_PARTIAL = "partial"
_IMAGE_STAGE_FINAL = "final"
_PARTIAL_IMAGE_SOURCE = "response.image_generation_call.partial_image"
_IMAGE_SIZE_RE = re.compile(r"^\s*(\d+)\s*[xX]\s*(\d+)\s*$")


def _to_text(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, list):
        parts: list[str] = []
        for item in value:
            text = _to_text(item).strip()
            if text != "":
                parts.append(text)
        return " ".join(parts)
    if isinstance(value, dict):
        for key in ("text", "value", "name", "description"):
            text = value.get(key)
            if isinstance(text, str) and text.strip() != "":
                return text
    return str(value)


def _normalize_positive_dimension(value: Any) -> Optional[int]:
    if isinstance(value, bool):
        return None
    if isinstance(value, int):
        return value if value > 0 else None
    if isinstance(value, float):
        as_int = int(value)
        return as_int if as_int > 0 else None
    if isinstance(value, str):
        stripped = value.strip()
        if stripped == "":
            return None
        if stripped.isdigit():
            parsed = int(stripped)
            return parsed if parsed > 0 else None
    return None


def _parse_image_dimensions_from_size(
    value: Any,
) -> tuple[Optional[int], Optional[int]]:
    if isinstance(value, str):
        match = _IMAGE_SIZE_RE.match(value)
        if match is None:
            return (None, None)
        width = int(match.group(1))
        height = int(match.group(2))
        return (
            width if width > 0 else None,
            height if height > 0 else None,
        )

    if isinstance(value, dict):
        return (
            _normalize_positive_dimension(value.get("width")),
            _normalize_positive_dimension(value.get("height")),
        )

    if isinstance(value, list) and len(value) >= 2:
        return (
            _normalize_positive_dimension(value[0]),
            _normalize_positive_dimension(value[1]),
        )

    return (None, None)


def _extract_image_dimensions(
    *,
    item: Optional[dict] = None,
    event: Optional[dict] = None,
) -> tuple[Optional[int], Optional[int]]:
    width = None
    height = None

    if isinstance(item, dict):
        width = _normalize_positive_dimension(item.get("width"))
        height = _normalize_positive_dimension(item.get("height"))

    if isinstance(event, dict):
        if width is None:
            width = _normalize_positive_dimension(event.get("width"))
        if height is None:
            height = _normalize_positive_dimension(event.get("height"))

    for source in (item, event):
        if not isinstance(source, dict):
            continue
        if width is None or height is None:
            parsed_width, parsed_height = _parse_image_dimensions_from_size(
                source.get("size")
            )
            if width is None and parsed_width is not None:
                width = parsed_width
            if height is None and parsed_height is not None:
                height = parsed_height

        if width is None or height is None:
            parsed_width, parsed_height = _parse_image_dimensions_from_size(
                source.get("resolution")
            )
            if width is None and parsed_width is not None:
                width = parsed_width
            if height is None and parsed_height is not None:
                height = parsed_height

    return (width, height)


def _image_status_from_state(*, state: str) -> str:
    normalized = state.strip().lower()
    if normalized in ("queued", "pending", "searching"):
        return "queued"
    if normalized in ("running", "in_progress"):
        return "generating"
    if normalized in ("completed",):
        return "completed"
    if normalized in ("failed",):
        return "failed"
    if normalized in ("cancelled",):
        return "cancelled"
    return "info"


def _first_nested_text(*, value: Any, keys: tuple[str, ...]) -> str:
    key_set = {key.lower() for key in keys}

    if isinstance(value, dict):
        for key, nested in value.items():
            if key.lower() in key_set:
                text = _to_text(nested).strip()
                if text != "":
                    return text

        for nested in value.values():
            text = _first_nested_text(value=nested, keys=keys)
            if text != "":
                return text

    elif isinstance(value, list):
        for nested in value:
            text = _first_nested_text(value=nested, keys=keys)
            if text != "":
                return text

    return ""


def _normalize_name(value: str) -> str:
    return "".join(ch for ch in value.lower() if ch.isalnum())


def _normalize_status_value(value: Any) -> Optional[str]:
    if not isinstance(value, str):
        return None

    normalized = value.strip().lower()
    if normalized == "":
        return None

    normalized = normalized.replace("-", "_").replace(" ", "_")
    if normalized == "inprogress":
        normalized = "in_progress"

    if normalized in ("failed", "error", "errored", "rejected"):
        return "failed"
    if normalized in ("cancelled", "canceled", "interrupted", "aborted", "stopped"):
        return "cancelled"
    if normalized in ("queued", "pending", "waiting"):
        return "queued"
    if normalized in ("running", "started", "starting", "in_progress"):
        return "in_progress"
    if normalized in (
        "completed",
        "complete",
        "done",
        "succeeded",
        "success",
        "finished",
    ):
        return "completed"

    if "fail" in normalized or "error" in normalized:
        return "failed"
    if "cancel" in normalized or "interrupt" in normalized or "abort" in normalized:
        return "cancelled"
    if "queue" in normalized or "pending" in normalized or "wait" in normalized:
        return "queued"
    if "complete" in normalized or "success" in normalized or "done" in normalized:
        return "completed"
    if "progress" in normalized or "running" in normalized or "start" in normalized:
        return "in_progress"

    return None


def _kind_from_item_type(*, item_type: str) -> Optional[str]:
    normalized = _normalize_name(item_type)
    if normalized == "":
        return None
    if normalized in (
        "mcpcall",
        "mcplisttools",
        "functioncall",
        "functioncalloutput",
        "customtoolcall",
        "hostedtoolcall",
        "toolcall",
        "computercall",
    ):
        return "tool"
    if normalized in ("websearchcall",):
        return "web"
    if normalized in ("filesearchcall",):
        return "search"
    if normalized in ("applypatchcall",):
        return "diff"
    if normalized in ("codeinterpretercall", "shellcall"):
        return "exec"
    if normalized in ("imagegenerationcall",):
        return "image"
    if normalized in ("reasoning",):
        return "reasoning"
    if normalized in ("message", "agentmessage"):
        return "message"
    return None


def _normalize_state_from_response_type(*, event_type: str) -> str:
    lower = event_type.lower()
    if lower.endswith(".failed"):
        return "failed"
    if lower.endswith(".cancelled") or lower.endswith(".canceled"):
        return "cancelled"
    if lower.endswith(".completed") or lower == "response.completed":
        return "completed"
    if (
        lower.endswith(".queued")
        or lower.endswith(".pending")
        or lower == "response.queued"
    ):
        return "queued"
    if (
        lower.endswith(".in_progress")
        or lower.endswith(".searching")
        or lower.endswith(".started")
        or lower.endswith(".generating")
        or lower.endswith(".added")
        or lower == "response.in_progress"
    ):
        return "in_progress"
    if lower.endswith(".done"):
        return "completed"
    return "info"


def _normalize_kind_from_response_type(*, event_type: str) -> str:
    lower = event_type.lower()
    if ".web_search_call." in lower:
        return "web"
    if ".file_search_call." in lower:
        return "search"
    if (
        ".mcp_call." in lower
        or ".mcp_list_tools." in lower
        or ".function_call." in lower
        or ".function_call_arguments." in lower
        or ".computer_call." in lower
    ):
        return "tool"
    if ".apply_patch_call." in lower:
        return "diff"
    if ".code_interpreter_call." in lower:
        return "exec"
    if ".image_generation_call." in lower:
        return "image"
    if lower.startswith("response.reasoning"):
        return "reasoning"
    if lower.startswith("response.output_item"):
        return "item"
    if lower.startswith("response."):
        return "turn"
    return "event"


def _response_identity(*, event: dict) -> tuple[Optional[str], Optional[str]]:
    item_id = event.get("item_id")
    if not isinstance(item_id, str) or item_id.strip() == "":
        item = event.get("item")
        if isinstance(item, dict):
            candidate = item.get("id")
            if isinstance(candidate, str) and candidate.strip() != "":
                item_id = candidate
            else:
                item_id = None
        else:
            item_id = None
    else:
        item_id = item_id.strip()

    response_id = None
    response = event.get("response")
    if isinstance(response, dict):
        candidate = response.get("id")
        if isinstance(candidate, str) and candidate.strip() != "":
            response_id = candidate.strip()

    return item_id, response_id


def _response_base_name(*, event_type: str) -> str:
    suffixes = (
        ".in_progress",
        ".searching",
        ".generating",
        ".completed",
        ".failed",
        ".cancelled",
        ".canceled",
        ".queued",
        ".pending",
        ".started",
    )
    for suffix in suffixes:
        if event_type.endswith(suffix):
            return event_type[: -len(suffix)]
    return event_type


def _computer_action_payload(*, event: dict) -> Optional[dict]:
    item = event.get("item")
    if isinstance(item, dict):
        action = item.get("action")
        if isinstance(action, dict):
            return action

    action = event.get("action")
    if isinstance(action, dict):
        return action

    return None


def _computer_action_type(*, event: dict) -> str:
    action = _computer_action_payload(event=event)
    if not isinstance(action, dict):
        return ""

    action_type = action.get("type")
    if not isinstance(action_type, str):
        return ""

    normalized = action_type.strip().lower().replace("-", "_").replace(" ", "_")
    return normalized


def _is_computer_call_event(*, event: dict) -> bool:
    event_name = event.get("name")
    if event_name == "computer.startup":
        return True

    item = event.get("item")
    if isinstance(item, dict):
        item_type = item.get("type")
        if isinstance(item_type, str) and _normalize_name(item_type) == "computercall":
            return True

    event_type = event.get("type")
    if isinstance(event_type, str) and ".computer_call." in event_type.lower():
        return True

    return False


def _computer_headline_for_state(*, event: dict, state: str) -> str:
    event_name = event.get("name")
    if event_name == "computer.startup":
        if state in _ACTIVE_STATES:
            return "Starting computer..."
        if state == "completed":
            return "Computer ready"
        if state == "failed":
            return "Failed to start computer"
        if state == "cancelled":
            return "Computer startup cancelled"
        return "Computer startup update"

    action_type = _computer_action_type(event=event)
    active_headlines = {
        "click": "Clicking on page",
        "double_click": "Double-clicking on page",
        "scroll": "Scrolling page",
        "type": "Typing text",
        "keypress": "Pressing keys",
        "move": "Moving cursor",
        "drag": "Dragging on page",
        "wait": "Waiting",
        "goto": "Opening page",
        "back": "Navigating back",
        "forward": "Navigating forward",
        "screenshot": "Capturing screenshot",
    }
    completed_headlines = {
        "click": "Clicked on page",
        "double_click": "Double-clicked on page",
        "scroll": "Scrolled page",
        "type": "Typed text",
        "keypress": "Pressed keys",
        "move": "Moved cursor",
        "drag": "Dragged on page",
        "wait": "Wait complete",
        "goto": "Opened page",
        "back": "Went back",
        "forward": "Went forward",
        "screenshot": "Captured screenshot",
    }

    if state in _ACTIVE_STATES:
        return active_headlines.get(action_type, "Using computer")
    if state == "completed":
        return completed_headlines.get(action_type, "Computer action complete")
    if state == "failed":
        return "Computer action failed"
    if state == "cancelled":
        return "Computer action cancelled"
    return "Computer action"


def _computer_scroll_direction_detail(*, action_payload: dict) -> Optional[str]:
    raw_scroll_x = action_payload.get("scroll_x")
    raw_scroll_y = action_payload.get("scroll_y")

    scroll_x = int(raw_scroll_x) if isinstance(raw_scroll_x, (int, float)) else 0
    scroll_y = int(raw_scroll_y) if isinstance(raw_scroll_y, (int, float)) else 0
    if scroll_x == 0 and scroll_y == 0:
        return None

    directions: list[str] = []
    if scroll_y != 0:
        directions.append("down" if scroll_y > 0 else "up")
    if scroll_x != 0:
        directions.append("right" if scroll_x > 0 else "left")

    if len(directions) == 1:
        return f"Direction: {directions[0]}"
    return f"Direction: {', '.join(directions)}"


def _headline_for_response_event(*, event: dict, kind: str, state: str) -> str:
    if kind == "turn":
        if state in _ACTIVE_STATES:
            return "Thinking"
        if state == "completed":
            return "Response Ready"
        if state == "failed":
            return "Response Failed"
        if state == "cancelled":
            return "Response Cancelled"
        return "Response Update"

    if kind == "web":
        if state in _ACTIVE_STATES:
            return "Searching Web"
        if state == "completed":
            return "Searched Web"
        if state == "failed":
            return "Web Search Failed"
        if state == "cancelled":
            return "Web Search Cancelled"
        return "Web Search"

    if kind == "search":
        if state in _ACTIVE_STATES:
            return "Searching Files"
        if state == "completed":
            return "Searched Files"
        if state == "failed":
            return "File Search Failed"
        if state == "cancelled":
            return "File Search Cancelled"
        return "File Search"

    if kind == "tool":
        if _is_computer_call_event(event=event):
            return _computer_headline_for_state(event=event, state=state)

        if state in _ACTIVE_STATES:
            return "Calling Tool"
        if state == "completed":
            return "Called Tool"
        if state == "failed":
            return "Tool Failed"
        if state == "cancelled":
            return "Tool Cancelled"
        return "Tool"

    if kind == "diff":
        if state in _ACTIVE_STATES:
            return "Applying Patch"
        if state == "completed":
            return "Applied Patch"
        if state == "failed":
            return "Patch Failed"
        if state == "cancelled":
            return "Patch Cancelled"
        return "Patch"

    if kind == "exec":
        if state in _ACTIVE_STATES:
            return "Running Command"
        if state == "completed":
            return "Ran Command"
        if state == "failed":
            return "Command Failed"
        if state == "cancelled":
            return "Command Cancelled"
        return "Command"

    if kind == "image":
        if state in _ACTIVE_STATES:
            return "Generating Image"
        if state == "completed":
            return "Generated Image"
        if state == "failed":
            return "Image Generation Failed"
        return "Image Generation"

    return "Event Update"


def _details_for_response_event(*, event: dict, kind: str) -> list[str]:
    details: list[str] = []
    seen: set[str] = set()

    def append_detail(text: str) -> None:
        normalized = " ".join(text.strip().lower().split())
        if normalized == "" or normalized in seen:
            return
        seen.add(normalized)
        details.append(text.strip())

    payload = event
    item = event.get("item")
    if isinstance(item, dict):
        payload = item

    action_payload = _computer_action_payload(event=event)
    action_type = _computer_action_type(event=event)

    if kind == "exec":
        command = _first_nested_text(
            value=payload,
            keys=("command", "cmd", "shell_command", "raw_command"),
        )
        if command != "":
            append_detail(command)

    if kind in ("web", "search"):
        query = _first_nested_text(value=payload, keys=("query", "pattern"))
        if query != "":
            append_detail(query)

    if kind == "tool":
        if _is_computer_call_event(event=event):
            if isinstance(action_payload, dict):
                if action_type == "scroll":
                    direction = _computer_scroll_direction_detail(
                        action_payload=action_payload
                    )
                    if isinstance(direction, str):
                        append_detail(direction)

                url = action_payload.get("url")
                if isinstance(url, str) and url.strip() != "":
                    append_detail(f"Opening: {url.strip()}")
        else:
            tool_name = _first_nested_text(
                value=payload,
                keys=("tool_name", "name", "server_label", "server", "tool"),
            )
            if tool_name != "":
                append_detail(f"Tool: {tool_name}")

    if kind == "diff":
        path = _first_nested_text(value=payload, keys=("path", "file", "filename"))
        if path != "":
            append_detail(path)

    return details


def _headline_with_primary_detail(
    *,
    headline: str,
    kind: str,
    details: list[str],
) -> str:
    if kind not in ("web", "search"):
        return headline

    if len(details) == 0:
        return headline

    detail = details[0].strip()
    if detail == "":
        return headline

    if detail.lower() in headline.lower():
        return headline

    max_chars = 120
    display = detail
    if len(display) > max_chars:
        display = display[:max_chars].rstrip() + "..."

    return f"{headline}: {display}"


def _item_id_for_message_event(*, event: dict) -> Optional[str]:
    item_id, _ = _response_identity(event=event)
    return item_id


def _content_index_for_message_event(*, event: dict) -> int:
    content_index = event.get("content_index")
    if isinstance(content_index, int) and content_index >= 0:
        return content_index
    return 0


def _message_text_from_content_part(*, part: Any) -> str:
    if not isinstance(part, dict):
        return ""

    part_type = part.get("type")
    if part_type == "output_text":
        text = part.get("text")
        if isinstance(text, str):
            return text
    elif part_type == "refusal":
        refusal = part.get("refusal")
        if isinstance(refusal, str):
            return refusal

    return ""


def _message_text_from_output_item(*, item: Any) -> str:
    if not isinstance(item, dict):
        return ""

    item_type = item.get("type")
    if item_type != "message":
        return ""

    content = item.get("content")
    if not isinstance(content, list):
        return ""

    parts: list[str] = []
    for part in content:
        text = _message_text_from_content_part(part=part)
        if text != "":
            parts.append(text)

    return "".join(parts)


def _mime_type_from_output_format(*, output_format: Optional[str]) -> str:
    if not isinstance(output_format, str):
        return "image/png"

    normalized = output_format.strip().lower().lstrip(".")
    if normalized == "":
        return "image/png"

    alias = {
        "jpg": "jpeg",
    }
    normalized = alias.get(normalized, normalized)
    return f"image/{normalized}"


def _image_bytes_from_output_item(*, item: dict) -> Optional[bytes]:
    if not isinstance(item, dict):
        return None

    for key in ("result", "image_base64", "image_b64", "b64_json", "data"):
        value = item.get(key)
        if isinstance(value, bytes):
            return value
        if isinstance(value, bytearray):
            return bytes(value)
        if isinstance(value, str) and value.strip() != "":
            try:
                return base64.b64decode(value)
            except Exception:
                logger.warning("unable to decode image payload from key '%s'", key)
                return None
        if isinstance(value, list):
            for entry in value:
                if isinstance(entry, str) and entry.strip() != "":
                    try:
                        return base64.b64decode(entry)
                    except Exception:
                        logger.warning(
                            "unable to decode image payload from key '%s' list entry",
                            key,
                        )
                        return None

    return None


def response_event_to_agent_event(event: dict) -> Optional[dict]:
    event_type = event.get("type")
    if not isinstance(event_type, str) or not event_type.startswith("response."):
        return None

    if event_type in _RESPONSE_NOISE_TYPES:
        return None
    if event_type in _RESPONSE_TURN_EVENTS:
        return None

    item = event.get("item")
    if not isinstance(item, dict):
        item = {}

    item_type = item.get("type")
    if not isinstance(item_type, str):
        item_type = ""

    state = _normalize_state_from_response_type(event_type=event_type)
    if state == "info":
        state = _normalize_status_value(item.get("status")) or "info"

    lower_event_type = event_type.lower()
    if state == "info" and lower_event_type == "response.output_item.added":
        state = "in_progress"
    elif state == "info" and lower_event_type == "response.output_item.done":
        state = "completed"

    if state == "info":
        return None

    kind = _kind_from_item_type(
        item_type=item_type
    ) or _normalize_kind_from_response_type(event_type=event_type)
    if kind in ("turn", "message", "reasoning", "item", "event"):
        return None
    if kind not in _SUPPORTED_EVENT_KINDS:
        return None

    item_id, response_id = _response_identity(event=event)
    base_name = _response_base_name(event_type=event_type)
    correlation_key = (
        f"item:{item_id}"
        if item_id is not None
        else (
            f"response:{base_name}:{response_id}"
            if response_id is not None
            else f"response:{base_name}"
        )
    )

    details = _details_for_response_event(event=event, kind=kind)
    headline = _headline_for_response_event(event=event, kind=kind, state=state)
    headline = _headline_with_primary_detail(
        headline=headline,
        kind=kind,
        details=details,
    )
    data = json.dumps(event, ensure_ascii=False, default=str)
    if len(data) > 8000:
        data = data[:8000] + "..."

    return {
        "type": "agent.event",
        "source": "openai",
        "name": event_type,
        "kind": kind,
        "state": state,
        "method": event_type,
        "correlation_key": correlation_key,
        "item_id": item_id,
        "item_type": None,
        "headline": headline,
        "details": details,
        "summary": headline,
        "data": data,
    }


class ResponsesThreadAdapter(ThreadAdapter):
    def __init__(
        self,
        *,
        image_captioner: Optional[ImageCaptioner] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._active_events_by_key: dict[str, Element] = {}
        self._images_db = ImagesDatabase(room=self._room)
        self._image_captioner = image_captioner
        self._saved_image_ids_by_item_id: dict[str, str] = {}
        self._saved_image_stage_by_item_id: dict[str, str] = {}

    async def stop(self) -> None:
        await super().stop()
        self._active_events_by_key.clear()
        self._saved_image_ids_by_item_id.clear()
        self._saved_image_stage_by_item_id.clear()

    async def handle_custom_event(
        self,
        *,
        messages: Element,
        event: dict,
    ) -> None:
        normalized_event = None

        event_type = event.get("type")
        if event_type in ("agent.event", "codex.event"):
            normalized_event = event
        else:
            normalized_event = response_event_to_agent_event(event)

        if not isinstance(normalized_event, dict):
            return

        source = normalized_event.get("source")
        if not isinstance(source, str) or source.strip() == "":
            source = "agent"
        source = source.strip()

        name = normalized_event.get("name")
        if not isinstance(name, str) or name.strip() == "":
            name = "agent.event"
        name = name.strip()

        kind = normalized_event.get("kind")
        if not isinstance(kind, str) or kind.strip() == "":
            return
        kind = kind.strip().lower()
        if kind not in _SUPPORTED_EVENT_KINDS:
            return

        state = normalized_event.get("state")
        if not isinstance(state, str) or state.strip() == "":
            state = "info"
        state = state.strip().lower()

        method = normalized_event.get("method")
        if not isinstance(method, str) or method.strip() == "":
            method = name
        method = method.strip()

        summary = normalized_event.get("summary")
        if not isinstance(summary, str) or summary.strip() == "":
            summary = method
        summary = summary.strip()

        headline = normalized_event.get("headline")
        if not isinstance(headline, str):
            headline = ""
        headline = headline.strip()

        item_id = normalized_event.get("item_id")
        if not isinstance(item_id, str):
            item_id = ""

        item_type = normalized_event.get("item_type")
        if not isinstance(item_type, str):
            item_type = ""
        event_path = normalized_event.get("path")
        if not isinstance(event_path, str):
            event_path = ""
        preview = normalized_event.get("preview")
        if not isinstance(preview, str):
            preview = ""

        details_value = normalized_event.get("details")
        if isinstance(details_value, list):
            lines = [line.strip() for line in details_value if isinstance(line, str)]
            details = "\n".join(line for line in lines if line != "")
        elif isinstance(details_value, str):
            details = details_value.strip()
        else:
            details = ""

        data = normalized_event.get("data")
        if not isinstance(data, str):
            data = json.dumps(normalized_event, ensure_ascii=False, default=str)
        persisted_data = data if kind == "diff" and data != "" else ""

        correlation_key = normalized_event.get("correlation_key")
        if not isinstance(correlation_key, str) or correlation_key.strip() == "":
            correlation_key = normalized_event.get("event_key")
        if not isinstance(correlation_key, str) or correlation_key.strip() == "":
            correlation_key = None
        else:
            correlation_key = correlation_key.strip()

        in_progress = state in _ACTIVE_STATES
        now = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")

        event_element: Element | None = None
        if correlation_key is not None:
            event_element = self._active_events_by_key.get(correlation_key)

        if event_element is None:
            attributes = {
                "id": str(uuid.uuid4()),
                "source": source,
                "name": name,
                "kind": kind,
                "state": state,
                "method": method,
                "item_id": item_id,
                "item_type": item_type,
                "path": event_path,
                "preview": preview,
                "summary": summary,
                "headline": headline,
                "details": details,
                "created_at": now,
                "updated_at": now,
            }
            if persisted_data != "":
                attributes["data"] = persisted_data
            event_element = messages.append_child(
                tag_name="event",
                attributes=attributes,
            )
        else:
            event_element.set_attribute("source", source)
            event_element.set_attribute("name", name)
            event_element.set_attribute("kind", kind)
            event_element.set_attribute("state", state)
            event_element.set_attribute("method", method)
            event_element.set_attribute("item_id", item_id)
            event_element.set_attribute("item_type", item_type)
            event_element.set_attribute("path", event_path)
            if preview != "" or event_element.get_attribute("preview") in (None, ""):
                event_element.set_attribute("preview", preview)
            event_element.set_attribute("summary", summary)
            event_element.set_attribute("headline", headline)
            if details != "" or event_element.get_attribute("details") in (None, ""):
                event_element.set_attribute("details", details)
            if persisted_data != "":
                event_element.set_attribute("data", persisted_data)
            event_element.set_attribute("updated_at", now)

        if correlation_key is not None:
            if in_progress:
                self._active_events_by_key[correlation_key] = event_element
            elif state in _TERMINAL_STATES:
                self._active_events_by_key.pop(correlation_key, None)

    async def _process_llm_events(self) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        doc_messages = None
        for prop in self._thread.root.get_children():
            if prop.tag_name == "messages":
                doc_messages = prop
                break

        if doc_messages is None:
            raise RoomException("messages element is missing from thread document")

        updates = asyncio.Queue()
        update_thread_stop = object()
        message_elements_by_key: dict[str, Element] = {}
        message_parts_by_key: dict[str, dict[int, str]] = {}
        latest_message_key: Optional[str] = None

        local_participant_name = self._room.local_participant.get_attribute("name")
        if not isinstance(local_participant_name, str):
            local_participant_name = ""

        # throttle updates so we don't send too many syncs over the wire at once
        async def update_thread() -> None:
            while True:
                entry = await updates.get()
                if entry is update_thread_stop:
                    break

                element, partial_text = entry
                changes: dict[Element, str] = {element: partial_text}
                should_stop = False
                while True:
                    try:
                        pending = updates.get_nowait()
                    except asyncio.QueueEmpty:
                        break

                    if pending is update_thread_stop:
                        should_stop = True
                        break

                    pending_element, pending_text = pending
                    changes[pending_element] = pending_text

                for changed_element, changed_text in changes.items():
                    changed_element["text"] = changed_text

                if should_stop:
                    break

        def resolve_message_key(*, event: dict) -> Optional[str]:
            item_id = _item_id_for_message_event(event=event)
            if item_id is not None:
                return f"item:{item_id}"
            return None

        def ensure_message_element(*, key: str, item_id: Optional[str]) -> Element:
            existing = message_elements_by_key.get(key)
            if existing is not None:
                return existing

            attributes: dict[str, str] = {
                "text": "",
                "created_at": datetime.now(timezone.utc)
                .isoformat()
                .replace("+00:00", "Z"),
                "author_name": local_participant_name,
            }
            if item_id is not None:
                attributes["id"] = item_id

            element = doc_messages.append_child(
                tag_name="message", attributes=attributes
            )
            message_elements_by_key[key] = element
            return element

        def queue_text_update(*, key: str, text: str) -> None:
            element = message_elements_by_key.get(key)
            if element is None:
                return
            updates.put_nowait((element, text))

        def set_part_text(*, key: str, content_index: int, text: str) -> None:
            by_index = message_parts_by_key.setdefault(key, {})
            by_index[content_index] = text
            merged = "".join(by_index[i] for i in sorted(by_index))
            queue_text_update(key=key, text=merged)

        update_thread_task = asyncio.create_task(update_thread())
        try:
            while True:
                evt = await self._llm_messages.get()

                event_type = evt.get("type")
                if not isinstance(event_type, str):
                    await self.handle_custom_event(messages=doc_messages, event=evt)
                    continue

                if event_type == "response.content_part.added":
                    part = evt.get("part")
                    part_type = part.get("type") if isinstance(part, dict) else None
                    if part_type == "reasoning_text":
                        continue

                    key = resolve_message_key(event=evt)
                    if key is None:
                        key = f"message:{uuid.uuid4()}"

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    part_text = _message_text_from_content_part(part=part)
                    if part_text != "":
                        set_part_text(
                            key=key,
                            content_index=_content_index_for_message_event(event=evt),
                            text=part_text,
                        )

                elif event_type == "response.output_text.delta":
                    key = resolve_message_key(event=evt) or latest_message_key
                    if key is None:
                        continue

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    content_index = _content_index_for_message_event(event=evt)
                    current = message_parts_by_key.setdefault(key, {}).get(
                        content_index, ""
                    )
                    delta = evt.get("delta")
                    if not isinstance(delta, str):
                        delta = ""
                    set_part_text(
                        key=key, content_index=content_index, text=current + delta
                    )

                elif event_type == "response.output_text.done":
                    key = resolve_message_key(event=evt) or latest_message_key
                    if key is None:
                        continue

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    text = evt.get("text")
                    if not isinstance(text, str):
                        text = ""
                    set_part_text(
                        key=key,
                        content_index=_content_index_for_message_event(event=evt),
                        text=text,
                    )
                    with tracer.start_as_current_span("chatbot.thread.message") as span:
                        span.set_attribute(
                            "from_participant_name", local_participant_name
                        )
                        span.set_attribute("role", "assistant")
                        span.set_attribute("text", text)

                elif event_type == "response.refusal.delta":
                    key = resolve_message_key(event=evt) or latest_message_key
                    if key is None:
                        continue

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    content_index = _content_index_for_message_event(event=evt)
                    current = message_parts_by_key.setdefault(key, {}).get(
                        content_index, ""
                    )
                    delta = evt.get("delta")
                    if not isinstance(delta, str):
                        delta = ""
                    set_part_text(
                        key=key, content_index=content_index, text=current + delta
                    )

                elif event_type == "response.refusal.done":
                    key = resolve_message_key(event=evt) or latest_message_key
                    if key is None:
                        continue

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    refusal = evt.get("refusal")
                    if not isinstance(refusal, str):
                        refusal = ""
                    set_part_text(
                        key=key,
                        content_index=_content_index_for_message_event(event=evt),
                        text=refusal,
                    )

                elif event_type == "response.content_part.done":
                    part = evt.get("part")
                    part_type = part.get("type") if isinstance(part, dict) else None
                    if part_type not in ("output_text", "refusal"):
                        continue

                    key = resolve_message_key(event=evt) or latest_message_key
                    if key is None:
                        continue

                    item_id = _item_id_for_message_event(event=evt)
                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    part_text = _message_text_from_content_part(part=part)
                    set_part_text(
                        key=key,
                        content_index=_content_index_for_message_event(event=evt),
                        text=part_text,
                    )

                elif event_type in (
                    "response.output_item.added",
                    "response.output_item.done",
                ):
                    item = evt.get("item")
                    item_id = _item_id_for_message_event(event=evt)
                    if not isinstance(item, dict):
                        await self.handle_custom_event(messages=doc_messages, event=evt)
                        continue

                    item_type = item.get("type")
                    if item_type == "image_generation_call":
                        if event_type == "response.output_item.done":
                            await self.handle_image_generation_output_item(
                                event=evt,
                                item=item,
                                source="response.output_item.done",
                            )
                        else:
                            await self.handle_image_generation_started(
                                event=evt,
                                item=item,
                            )
                        continue

                    if item_type != "message":
                        await self.handle_custom_event(messages=doc_messages, event=evt)
                        continue

                    if item_id is None:
                        key = latest_message_key
                    else:
                        key = f"item:{item_id}"

                    if key is None:
                        continue

                    ensure_message_element(key=key, item_id=item_id)
                    latest_message_key = key

                    message_text = _message_text_from_output_item(item=item)
                    if message_text != "":
                        set_part_text(key=key, content_index=0, text=message_text)

                elif event_type == "response.image_generation_call.partial_image":
                    await self.handle_image_generation_partial(event=evt)

                elif event_type == "meshagent.handler.added":
                    item = evt.get("item")
                    if not isinstance(item, dict):
                        continue

                    item_type = item.get("type")
                    if item_type == "shell_call":
                        await self.handle_shell_call_output(item=item)

                    elif item_type == "local_shell_call":
                        await self.handle_local_shell_call_output(item=item)

                    elif item_type == "image_generation_call":
                        await self.handle_image_generation_output_item(
                            event=evt,
                            item=item,
                            source="meshagent.handler.added",
                        )

                else:
                    await self.handle_custom_event(messages=doc_messages, event=evt)

        except asyncio.QueueShutDown:
            pass
        finally:
            updates.put_nowait(update_thread_stop)

        await update_thread_task

    def _messages_element(self) -> Element:
        if self._thread is None:
            raise RoomException("thread was not opened")

        for prop in self._thread.root.get_children():
            if prop.tag_name == "messages":
                return prop

        raise RoomException("messages element is missing from thread document")

    def _resolve_image_item_id(
        self,
        *,
        event: dict,
        item: Optional[dict] = None,
    ) -> str:
        if isinstance(item, dict):
            item_id = item.get("id")
            if isinstance(item_id, str) and item_id.strip() != "":
                return item_id

        event_item_id = event.get("item_id")
        if isinstance(event_item_id, str) and event_item_id.strip() != "":
            return event_item_id

        return str(uuid.uuid4())

    def _upsert_image_status(
        self,
        *,
        item_id: str,
        state: str,
        headline: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> None:
        created_by = self._room.local_participant.get_attribute("name")
        if not isinstance(created_by, str):
            created_by = ""

        self.write_image(
            message_id=item_id,
            width=width,
            height=height,
            created_by=created_by,
            status=_image_status_from_state(state=state),
            status_detail=headline,
        )

    async def _emit_image_status_event(
        self,
        *,
        messages: Element,
        item_id: str,
        state: str,
        headline: str,
        details: Optional[list[str]] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> None:
        normalized_details = [line for line in (details or []) if isinstance(line, str)]
        await self.handle_custom_event(
            messages=messages,
            event={
                "type": "agent.event",
                "source": "openai",
                "name": "response.image_generation_call.partial_image",
                "kind": "image",
                "state": state,
                "method": "response.image_generation_call.partial_image",
                # Match response_event_to_agent_event correlation semantics so
                # custom image status updates mutate the existing image event line.
                "correlation_key": f"item:{item_id}",
                "item_id": item_id,
                "item_type": "image_generation_call",
                "summary": headline,
                "headline": headline,
                "details": normalized_details,
                "data": json.dumps(
                    {
                        "item_id": item_id,
                        "state": state,
                        "headline": headline,
                        "details": normalized_details,
                        "width": width,
                        "height": height,
                    }
                ),
            },
        )
        self._upsert_image_status(
            item_id=item_id,
            state=state,
            headline=headline,
            width=width,
            height=height,
        )

    async def _persist_generated_image(
        self,
        *,
        item_id: str,
        image_bytes: bytes,
        mime_type: str,
        created_by: str,
        source: str,
        annotations: Optional[dict[str, str]] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> None:
        is_partial_source = source == _PARTIAL_IMAGE_SOURCE
        incoming_stage = (
            _IMAGE_STAGE_PARTIAL if is_partial_source else _IMAGE_STAGE_FINAL
        )

        existing_stage = self._saved_image_stage_by_item_id.get(item_id)
        if existing_stage == _IMAGE_STAGE_FINAL:
            return
        if (
            existing_stage == _IMAGE_STAGE_PARTIAL
            and incoming_stage == _IMAGE_STAGE_PARTIAL
        ):
            return

        messages = self._messages_element()
        progress_headline = (
            "Saving final image"
            if existing_stage == _IMAGE_STAGE_PARTIAL
            and incoming_stage == _IMAGE_STAGE_FINAL
            else "Saving image"
        )
        await self._emit_image_status_event(
            messages=messages,
            item_id=item_id,
            state="in_progress",
            headline=progress_headline,
            width=width,
            height=height,
        )

        payload_annotations = {
            "source": source,
            "item_id": item_id,
            "stage": incoming_stage,
        }
        if isinstance(annotations, dict):
            for key, value in annotations.items():
                payload_annotations[str(key)] = str(value)
        if width is not None:
            payload_annotations["width"] = str(width)
        if height is not None:
            payload_annotations["height"] = str(height)

        try:
            saved_image = await asyncio.wait_for(
                self._images_db.save(
                    data=image_bytes,
                    mime_type=mime_type,
                    created_by=created_by,
                    annotations=payload_annotations,
                ),
                timeout=_IMAGE_DB_SAVE_TIMEOUT_SECONDS,
            )
        except Exception as ex:
            logger.error("failed to save generated image to database", exc_info=ex)
            await self._emit_image_status_event(
                messages=messages,
                item_id=item_id,
                state="failed",
                headline="Image save failed",
                details=[str(ex)],
                width=width,
                height=height,
            )
            return

        try:
            self.write_image(
                message_id=item_id,
                image_id=saved_image.id,
                mime_type=saved_image.mime_type,
                created_at=saved_image.created_at,
                created_by=saved_image.created_by,
                width=width,
                height=height,
                status=_image_status_from_state(state="completed"),
                status_detail="Image saved",
            )
        except Exception as ex:
            logger.error("failed to attach saved image to thread", exc_info=ex)
            await self._emit_image_status_event(
                messages=messages,
                item_id=item_id,
                state="failed",
                headline="Image attach failed",
                details=[str(ex)],
                width=width,
                height=height,
            )
            return

        if incoming_stage == _IMAGE_STAGE_PARTIAL:
            await self._emit_image_status_event(
                messages=messages,
                item_id=item_id,
                state="completed",
                headline="Image saved",
                width=width,
                height=height,
            )
        else:
            await self._emit_image_status_event(
                messages=messages,
                item_id=item_id,
                state="completed",
                headline=(
                    "Final image saved"
                    if existing_stage == _IMAGE_STAGE_PARTIAL
                    else "Image saved"
                ),
                width=width,
                height=height,
            )

        self._saved_image_ids_by_item_id[item_id] = saved_image.id
        self._saved_image_stage_by_item_id[item_id] = incoming_stage
        logger.info("Saved generated image %s to images database", saved_image.id)

    async def handle_image_generation_started(
        self,
        *,
        event: dict,
        item: dict,
    ) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        if not isinstance(item, dict):
            return

        item_id = self._resolve_image_item_id(event=event, item=item)
        width, height = _extract_image_dimensions(item=item, event=event)

        normalized_state = _normalize_status_value(item.get("status")) or "in_progress"
        if normalized_state == "failed":
            headline = "Image generation failed"
        elif normalized_state == "cancelled":
            headline = "Image generation cancelled"
        elif normalized_state == "queued":
            headline = "Image generation queued"
        else:
            normalized_state = "in_progress"
            headline = "Generating image"

        messages = self._messages_element()
        await self._emit_image_status_event(
            messages=messages,
            item_id=item_id,
            state=normalized_state,
            headline=headline,
            width=width,
            height=height,
        )

    async def handle_image_generation_output_item(
        self,
        *,
        event: dict,
        item: dict,
        source: str,
    ) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        if not isinstance(item, dict):
            return

        item_id = self._resolve_image_item_id(event=event, item=item)
        width, height = _extract_image_dimensions(item=item, event=event)

        image_bytes = _image_bytes_from_output_item(item=item)
        if image_bytes is None:
            messages = self._messages_element()
            normalized_state = _normalize_status_value(
                item.get("status")
            ) or _normalize_status_value(event.get("status"))

            if normalized_state in ("failed", "cancelled"):
                details: list[str] = []
                detail = _first_nested_text(
                    value=item,
                    keys=("error", "errors", "message", "reason", "detail"),
                ).strip()
                if detail != "":
                    details.append(detail)
                await self._emit_image_status_event(
                    messages=messages,
                    item_id=item_id,
                    state=normalized_state,
                    headline=(
                        "Image generation failed"
                        if normalized_state == "failed"
                        else "Image generation cancelled"
                    ),
                    details=details,
                    width=width,
                    height=height,
                )
            elif normalized_state in ("queued", "in_progress"):
                await self._emit_image_status_event(
                    messages=messages,
                    item_id=item_id,
                    state=normalized_state,
                    headline=(
                        "Image generation queued"
                        if normalized_state == "queued"
                        else "Generating image"
                    ),
                    width=width,
                    height=height,
                )
            return

        output_format = item.get("output_format")
        if not isinstance(output_format, str) or output_format.strip() == "":
            output_format = event.get("output_format")
        mime_type = _mime_type_from_output_format(output_format=output_format)

        created_by = self._room.local_participant.get_attribute("name")
        if not isinstance(created_by, str):
            created_by = ""

        item_annotations: dict[str, str] = {}
        for key in ("status", "size", "quality", "background", "output_format"):
            value = item.get(key)
            if isinstance(value, str) and value.strip() != "":
                item_annotations[key] = value.strip()

        await self._persist_generated_image(
            item_id=item_id,
            image_bytes=image_bytes,
            mime_type=mime_type,
            created_by=created_by,
            source=source,
            annotations=item_annotations,
            width=width,
            height=height,
        )

    async def handle_image_generation_partial(
        self,
        *,
        event: dict,
    ) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        item_id = self._resolve_image_item_id(event=event)
        width, height = _extract_image_dimensions(event=event)

        partial_image_b64 = event.get("partial_image_b64")
        if not isinstance(partial_image_b64, str) or partial_image_b64.strip() == "":
            raise RoomException("missing partial_image_b64 in image generation event")

        try:
            image_bytes = base64.b64decode(partial_image_b64)
        except Exception as ex:
            messages = self._messages_element()
            await self._emit_image_status_event(
                messages=messages,
                item_id=item_id,
                state="failed",
                headline="Image save failed",
                details=[f"invalid image payload: {ex}"],
                width=width,
                height=height,
            )
            return

        output_format = event.get("output_format")
        mime_type = _mime_type_from_output_format(output_format=output_format)

        created_by = self._room.local_participant.get_attribute("name")
        if not isinstance(created_by, str):
            created_by = ""

        await self._persist_generated_image(
            item_id=item_id,
            image_bytes=image_bytes,
            mime_type=mime_type,
            created_by=created_by,
            source=_PARTIAL_IMAGE_SOURCE,
            annotations={
                "output_format": str(output_format)
                if isinstance(output_format, str)
                else "",
            },
            width=width,
            height=height,
        )

    async def handle_local_shell_call_output(
        self,
        *,
        item: dict,
    ) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages = self._thread.root.get_children_by_tag_name("messages")[0]

        action = item["action"]
        command = action["command"]
        working_dir = action["working_dir"]

        for prop in self._thread.root.get_children():
            if prop.tag_name == "messages":
                messages = prop
                break

        exec_element = messages.append_child(
            tag_name="exec",
            attributes={"command": shlex.join(command), "pwd": working_dir},
        )

        evt = await self._llm_messages.get()

        if evt["type"] != "meshagent.handler.done":
            raise RoomException("expected meshagent.handler.done")

        error = evt.get("error")
        item = evt.get("item")

        if error is not None:
            pass

        if item is not None:
            if item["type"] != "local_shell_call_output":
                raise RoomException("expected local_shell_call_output")

            exec_element.set_attribute("result", item["output"])

    async def handle_shell_call_output(
        self,
        *,
        item: dict,
    ) -> None:
        if self._thread is None:
            raise RoomException("thread was not opened")

        messages = self._thread.root.get_children_by_tag_name("messages")[0]

        action = item["action"]
        commands = action["commands"]

        exec_elements = []
        for command in commands:
            exec_element = messages.append_child(
                tag_name="exec",
                attributes={"command": command},
            )
            exec_elements.append(exec_element)

        evt = await self._llm_messages.get()

        if evt["type"] != "meshagent.handler.done":
            raise RoomException("expected meshagent.handler.done")

        error = evt.get("error")
        item = evt.get("item")

        if error is not None:
            pass

        if item is not None:
            if item["type"] != "shell_call_output":
                raise RoomException("expected shell_call_output")

            results = item["output"]

            for i in range(0, len(results)):
                result = results[i]
                exec_element = exec_elements[i]
                if "exit_code" in result["outcome"]:
                    exec_element.set_attribute(
                        "exit_code", result["outcome"]["exit_code"]
                    )

                exec_element.set_attribute("outcome", result["outcome"]["type"])
                exec_element.set_attribute("stdout", result["stdout"])
                exec_element.set_attribute("stderr", result["stderr"])
